# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

import pytest

from . import autoconfig

from pygccxml import utils
from pygccxml import parser
from pygccxml import declarations


def is_sub_path(root, some_path):
    root = utils.normalize_path(root)
    some_path = utils.normalize_path(some_path)
    return some_path.startswith(root)


TEST_FILES = [
    "core_ns_join_1.hpp",
    "core_ns_join_2.hpp",
    "core_ns_join_3.hpp",
    "core_membership.hpp",
    "core_class_hierarchy.hpp",
    "core_types.hpp",
    "core_diamand_hierarchy_base.hpp",
    "core_diamand_hierarchy_derived1.hpp",
    "core_diamand_hierarchy_derived2.hpp",
    "core_diamand_hierarchy_final_derived.hpp",
    "core_overloads_1.hpp",
    "core_overloads_2.hpp",
    "abstract_classes.hpp",
]


@pytest.fixture
def global_ns_fixture_all_at_once1():
    COMPILATION_MODE = parser.COMPILATION_MODE.ALL_AT_ONCE
    INIT_OPTIMIZER = True
    config = autoconfig.cxx_parsers_cfg.config.clone()
    decls = parser.parse(TEST_FILES, config, COMPILATION_MODE)
    global_ns = declarations.get_global_namespace(decls)
    if INIT_OPTIMIZER:
        global_ns.init_optimizer()
    return global_ns


@pytest.fixture
def global_ns_fixture_all_at_once2():
    COMPILATION_MODE = parser.COMPILATION_MODE.ALL_AT_ONCE
    INIT_OPTIMIZER = False
    config = autoconfig.cxx_parsers_cfg.config.clone()
    decls = parser.parse(TEST_FILES, config, COMPILATION_MODE)
    global_ns = declarations.get_global_namespace(decls)
    if INIT_OPTIMIZER:
        global_ns.init_optimizer()
    return global_ns


@pytest.fixture
def global_ns_fixture_file_by_file1():
    COMPILATION_MODE = parser.COMPILATION_MODE.FILE_BY_FILE
    INIT_OPTIMIZER = True
    config = autoconfig.cxx_parsers_cfg.config.clone()
    decls = parser.parse(TEST_FILES, config, COMPILATION_MODE)
    global_ns = declarations.get_global_namespace(decls)
    if INIT_OPTIMIZER:
        global_ns.init_optimizer()
    return global_ns


@pytest.fixture
def global_ns_fixture_file_by_file2():
    COMPILATION_MODE = parser.COMPILATION_MODE.FILE_BY_FILE
    INIT_OPTIMIZER = False
    config = autoconfig.cxx_parsers_cfg.config.clone()
    decls = parser.parse(TEST_FILES, config, COMPILATION_MODE)
    global_ns = declarations.get_global_namespace(decls)
    if INIT_OPTIMIZER:
        global_ns.init_optimizer()
    return global_ns


@pytest.fixture
def global_ns(request):
    return request.getfixturevalue(request.param)


@pytest.mark.parametrize(
    "global_ns",
    [
        "global_ns_fixture_all_at_once1",
        "global_ns_fixture_all_at_once2",
        "global_ns_fixture_file_by_file1",
        "global_ns_fixture_file_by_file2",
    ],
    indirect=True,
)
def test_top_parent(global_ns):
    enum = global_ns.enumeration("::ns::ns32::E33")
    assert global_ns is enum.top_parent


# tests namespaces join functionality. described in gccxml.py
@pytest.mark.parametrize(
    "global_ns",
    [
        "global_ns_fixture_all_at_once1",
        "global_ns_fixture_all_at_once2",
        "global_ns_fixture_file_by_file1",
        "global_ns_fixture_file_by_file2",
    ],
    indirect=True,
)
def test_nss_join(global_ns):
    # list of all namespaces
    nss = ["::ns", "::ns::ns12", "::ns::ns22", "::ns::ns32"]
    # list of all namespaces that have unnamed namespace
    # unnamed_nss = nss[1:] doing nothing with this list ?
    # list of all enums [0:2] [3:5] [6:8] - has same parent
    enums = [
        "::E11",
        "::E21",
        "::E31",
        "::ns::E12",
        "::ns::E22",
        "::ns::E32",
        "::ns::ns12::E13",
        "::ns::ns22::E23",
        "::ns::ns32::E33",
    ]

    for ns in nss:
        global_ns.namespace(ns)

    for enum in enums:
        global_ns.enumeration(enum)

    ns = global_ns.namespace(nss[0])
    ns12 = global_ns.namespace(nss[1])
    ns22 = global_ns.namespace(nss[2])
    ns32 = global_ns.namespace(nss[3])
    assert ns is ns12.parent is ns22.parent is ns32.parent

    e11 = global_ns.enumeration(enums[0])
    e21 = global_ns.enumeration(enums[1])
    e31 = global_ns.enumeration(enums[2])
    assert e11.parent is e21.parent is e31.parent

    nse12 = global_ns.enumeration(enums[3])
    nse23 = global_ns.enumeration(enums[4])
    nse33 = global_ns.enumeration(enums[5])
    assert ns is nse12.parent is nse23.parent is nse33.parent


def _test_ns_membership(ns, enum_name):
    unnamed_enum = ns.enumeration(
        lambda d: d.name == ""
        and is_sub_path(autoconfig.data_directory, d.location.file_name),
        recursive=False,
    )
    assert unnamed_enum in ns.declarations

    enum = ns.enumeration(enum_name, recursive=False)
    assert enum in ns.declarations
    assert unnamed_enum.parent is ns
    assert enum.parent is ns


def _test_class_membership(class_inst, enum_name, access):
    # getting enum through get_members function
    nested_enum1 = class_inst.enumeration(
        name=enum_name, function=declarations.access_type_matcher_t(access)
    )

    # getting enum through declarations property
    nested_enum2 = class_inst.enumeration(enum_name)

    # it shoud be same object
    assert nested_enum1 is nested_enum2

    # check whether we meaning same class instance
    assert class_inst is nested_enum1.parent is nested_enum2.parent


@pytest.mark.parametrize(
    "global_ns",
    [
        "global_ns_fixture_all_at_once1",
        "global_ns_fixture_all_at_once2",
        "global_ns_fixture_file_by_file1",
        "global_ns_fixture_file_by_file2",
    ],
    indirect=True,
)
# test gccxml_file_reader_t._update_membership algorithm
def test_membership(global_ns):
    core_membership = global_ns.namespace("membership")
    _test_ns_membership(global_ns, "EGlobal")
    _test_ns_membership(core_membership.namespace("enums_ns"), "EWithin")
    _test_ns_membership(core_membership.namespace(""), "EWithinUnnamed")
    class_nested_enums = core_membership.class_("class_for_nested_enums_t")
    _test_class_membership(
        class_nested_enums,
        "ENestedPublic",
        declarations.ACCESS_TYPES.PUBLIC
    )
    _test_class_membership(
        class_nested_enums,
        "ENestedProtected",
        declarations.ACCESS_TYPES.PROTECTED
    )
    _test_class_membership(
        class_nested_enums,
        "ENestedPrivate",
        declarations.ACCESS_TYPES.PRIVATE
    )


@pytest.mark.parametrize(
    "global_ns",
    [
        "global_ns_fixture_all_at_once1",
        "global_ns_fixture_all_at_once2",
        "global_ns_fixture_file_by_file1",
        "global_ns_fixture_file_by_file2",
    ],
    indirect=True,
)
def test_mangled_name_namespace(global_ns):
    std = global_ns.namespace("std")
    assert std is not None
    assert std.mangled is None


@pytest.mark.parametrize(
    "global_ns",
    [
        "global_ns_fixture_all_at_once1",
        "global_ns_fixture_all_at_once2",
        "global_ns_fixture_file_by_file1",
        "global_ns_fixture_file_by_file2",
    ],
    indirect=True,
)
def test_mangled_name_functions(global_ns):
    # This works with gccxml and castxml
    ns = global_ns.namespace("overloads")
    do_nothing = ns.calldefs("do_nothing", recursive=False)
    assert do_nothing.mangled is not None


@pytest.mark.parametrize(
    "global_ns",
    [
        "global_ns_fixture_all_at_once1",
        "global_ns_fixture_all_at_once2",
        "global_ns_fixture_file_by_file1",
        "global_ns_fixture_file_by_file2",
    ],
    indirect=True,
)
def test_mangled_name_variable(global_ns):
    # This works with gccxml and castxml
    var_inst = global_ns.variable("array255")
    assert var_inst.mangled is not None


def _test_is_based_and_derived(base, derived, access):
    dhi_v = declarations.hierarchy_info_t(derived, access, True)
    dhi_not_v = declarations.hierarchy_info_t(derived, access, False)
    assert dhi_v in base.derived or dhi_not_v in base.derived

    bhi_v = declarations.hierarchy_info_t(base, access, True)
    bhi_not_v = declarations.hierarchy_info_t(base, access, False)

    assert bhi_v in derived.bases or bhi_not_v in derived.bases


@pytest.mark.parametrize(
    "global_ns",
    [
        "global_ns_fixture_all_at_once1",
        "global_ns_fixture_all_at_once2",
        "global_ns_fixture_file_by_file1",
        "global_ns_fixture_file_by_file2",
    ],
    indirect=True,
)
def test_class_hierarchy(global_ns):
    class_hierarchy = global_ns.namespace("class_hierarchy")

    base = class_hierarchy.class_("base_t")
    other_base = class_hierarchy.class_("other_base_t")
    derived_public = class_hierarchy.class_("derived_public_t")
    derived_protected = class_hierarchy.class_("derived_protected_t")
    derived_private = class_hierarchy.class_("derived_private_t")
    multi_derived = class_hierarchy.class_("multi_derived_t")

    _test_is_based_and_derived(
        base,
        derived_public,
        declarations.ACCESS_TYPES.PUBLIC
    )
    _test_is_based_and_derived(
        base,
        derived_protected,
        declarations.ACCESS_TYPES.PROTECTED
    )
    _test_is_based_and_derived(
        base,
        derived_private,
        declarations.ACCESS_TYPES.PRIVATE
    )
    _test_is_based_and_derived(
        base, multi_derived,
        declarations.ACCESS_TYPES.PROTECTED
    )
    _test_is_based_and_derived(
        other_base,
        multi_derived,
        declarations.ACCESS_TYPES.PRIVATE
    )
    _test_is_based_and_derived(
        derived_private,
        multi_derived,
        declarations.ACCESS_TYPES.PRIVATE
    )


def _test_is_same_bases(derived1, derived2):
    bases1 = set(
        [id(hierarchy_info.related_class) for hierarchy_info in derived1.bases]
    )
    bases2 = set(
        [id(hierarchy_info.related_class) for hierarchy_info in derived2.bases]
    )
    assert bases1 == bases2


@pytest.mark.parametrize(
    "global_ns",
    [
        "global_ns_fixture_all_at_once1",
        "global_ns_fixture_all_at_once2",
        "global_ns_fixture_file_by_file1",
        "global_ns_fixture_file_by_file2",
    ],
    indirect=True,
)
def test_class_join(global_ns):
    diamand_hierarchy = global_ns.namespace("diamand_hierarchy")
    base = diamand_hierarchy.class_("base_t")
    derived1 = diamand_hierarchy.class_("derived1_t")
    derived2 = diamand_hierarchy.class_("derived2_t")
    final_derived = diamand_hierarchy.class_("final_derived_t")

    _test_is_based_and_derived(
        base,
        derived1,
        declarations.ACCESS_TYPES.PUBLIC
    )
    _test_is_based_and_derived(
        base,
        derived2,
        declarations.ACCESS_TYPES.PUBLIC
    )
    _test_is_based_and_derived(
        derived1,
        final_derived,
        declarations.ACCESS_TYPES.PUBLIC
    )
    _test_is_based_and_derived(
        derived2,
        final_derived,
        declarations.ACCESS_TYPES.PUBLIC
    )
    _test_is_same_bases(derived1, derived2)


@pytest.mark.parametrize(
    "global_ns",
    [
        "global_ns_fixture_all_at_once1",
        "global_ns_fixture_all_at_once2",
        "global_ns_fixture_file_by_file1",
        "global_ns_fixture_file_by_file2",
    ],
    indirect=True,
)
def test_fundamental_types(global_ns):
    # check whether all build in types could be constructed
    errors = []
    for (
        fundamental_type_name,
        fundamental_type,
    ) in declarations.FUNDAMENTAL_TYPES.items():
        if "complex" in fundamental_type_name:
            continue  # I check this in an other tester
        if isinstance(
            fundamental_type, (declarations.int128_t, declarations.uint128_t)
        ):
            continue  # I don't have test case for this
        if isinstance(fundamental_type, declarations.java_fundamental_t):
            continue  # I don't check this at all
        typedef_name = "typedef_" + fundamental_type_name.replace(" ", "_")
        typedef = global_ns.decl(
            decl_type=declarations.typedef_t,
            name=typedef_name
        )
        assert typedef is not None
        if typedef.decl_type.decl_string != fundamental_type.decl_string:
            errors.append(
                "there is a difference between typedef base type "
                + "name('%s') and expected one('%s')"
                % (typedef.decl_type.decl_string, fundamental_type.decl_string)
            )
    assert len(errors) == 0


@pytest.mark.parametrize(
    "global_ns",
    [
        "global_ns_fixture_all_at_once1",
        "global_ns_fixture_all_at_once2",
        "global_ns_fixture_file_by_file1",
        "global_ns_fixture_file_by_file2",
    ],
    indirect=True,
)
def test_compound_types(global_ns, helpers):
    typedef_inst = global_ns.decl(
        decl_type=declarations.typedef_t, name="typedef_const_int"
    )
    helpers._test_type_composition(
        typedef_inst.decl_type, declarations.const_t, declarations.int_t
    )

    typedef_inst = global_ns.decl(
        decl_type=declarations.typedef_t, name="typedef_pointer_int"
    )
    helpers._test_type_composition(
        typedef_inst.decl_type, declarations.pointer_t, declarations.int_t
    )

    typedef_inst = global_ns.decl(
        decl_type=declarations.typedef_t, name="typedef_reference_int"
    )
    helpers._test_type_composition(
        typedef_inst.decl_type, declarations.reference_t, declarations.int_t
    )

    typedef_inst = global_ns.decl(
        decl_type=declarations.typedef_t,
        name="typedef_const_unsigned_int_const_pointer",
    )
    helpers._test_type_composition(
        typedef_inst.decl_type,
        declarations.const_t,
        declarations.pointer_t
    )
    helpers._test_type_composition(
        typedef_inst.decl_type.base,
        declarations.pointer_t,
        declarations.const_t
    )
    helpers._test_type_composition(
        typedef_inst.decl_type.base.base,
        declarations.const_t,
        declarations.unsigned_int_t,
    )

    typedef_inst = global_ns.decl(
        decl_type=declarations.typedef_t,
        name="typedef_volatile_int"
    )
    helpers._test_type_composition(
        typedef_inst.decl_type,
        declarations.volatile_t,
        declarations.int_t
    )

    var_inst = global_ns.variable("array255")
    helpers._test_type_composition(
        var_inst.decl_type,
        declarations.array_t,
        declarations.int_t
    )

    typedef_inst = global_ns.decl(
        decl_type=declarations.typedef_t,
        name="typedef_EFavoriteDrinks"
    )
    assert isinstance(typedef_inst.decl_type, declarations.declarated_t)
    enum_declaration = global_ns.enumeration("EFavoriteDrinks")
    assert typedef_inst.decl_type.declaration is enum_declaration


@pytest.mark.parametrize(
    "global_ns",
    [
        "global_ns_fixture_all_at_once1",
        "global_ns_fixture_all_at_once2",
        "global_ns_fixture_file_by_file1",
        "global_ns_fixture_file_by_file2",
    ],
    indirect=True,
)
def test_free_function_type(global_ns, helpers):
    function_ptr = global_ns.decl(
        decl_type=declarations.typedef_t,
        name="function_ptr")
    helpers._test_type_composition(
        function_ptr.decl_type,
        declarations.pointer_t,
        declarations.free_function_type_t,
    )
    function_type = function_ptr.decl_type.base
    assert isinstance(function_type.return_type, declarations.int_t)
    assert len(function_type.arguments_types) == 2
    assert isinstance(function_type.arguments_types[0], declarations.int_t)
    assert isinstance(function_type.arguments_types[1], declarations.double_t)


@pytest.mark.parametrize(
        "global_ns",
        ["global_ns_fixture_all_at_once1"],
        indirect=True
    )
def test_member_function_type(global_ns, helpers):
    function_ptr = global_ns.decl(
        decl_type=declarations.typedef_t, name="member_function_ptr_t"
    )
    helpers._test_type_composition(
        function_ptr.decl_type,
        declarations.pointer_t,
        declarations.member_function_type_t,
    )

    function_type = function_ptr.decl_type.base

    members_pointers = global_ns.class_("members_pointers_t")
    assert function_type.class_inst.declaration is members_pointers

    assert isinstance(function_type.return_type, declarations.int_t)
    assert len(function_type.arguments_types) == 1
    assert isinstance(function_type.arguments_types[0], declarations.double_t)
    assert function_type.has_const is True


@pytest.mark.parametrize(
    "global_ns",
    [
        "global_ns_fixture_all_at_once1",
        "global_ns_fixture_all_at_once2",
        "global_ns_fixture_file_by_file1",
        "global_ns_fixture_file_by_file2",
    ],
    indirect=True,
)
def test_member_variable_type(global_ns, helpers):
    mv = global_ns.decl(
        decl_type=declarations.typedef_t,
        name="member_variable_ptr_t"
    )
    helpers._test_type_composition(
        mv.decl_type,
        declarations.pointer_t,
        declarations.member_variable_type_t
    )

    members_pointers = global_ns.class_("members_pointers_t")
    assert members_pointers is not None
    helpers._test_type_composition(
        mv.decl_type.base,
        declarations.member_variable_type_t,
        declarations.declarated_t,
    )
    mv_type = mv.decl_type.base
    assert mv_type.base.declaration == members_pointers


@pytest.mark.parametrize(
    "global_ns",
    [
        "global_ns_fixture_all_at_once1",
        "global_ns_fixture_all_at_once2",
        "global_ns_fixture_file_by_file1",
        "global_ns_fixture_file_by_file2",
    ],
    indirect=True,
)
def test_overloading(global_ns):
    ns = global_ns.namespace("overloads")

    do_nothings = ns.calldefs("do_nothing", recursive=False)
    assert 4 == len(do_nothings)
    for index, do_nothing in enumerate(do_nothings):
        others = do_nothings[:index] + do_nothings[index + 1:]
        if set(do_nothing.overloads) != set(others):
            print("\nexisting: ")
            for x in do_nothing.overloads:
                print(str(x))
            print("\nexpected: ")
            for x in others:
                print(str(x))

        assert set(do_nothing.overloads) == set(others)


@pytest.mark.parametrize(
    "global_ns",
    [
        "global_ns_fixture_all_at_once1",
        "global_ns_fixture_all_at_once2",
        "global_ns_fixture_file_by_file1",
        "global_ns_fixture_file_by_file2",
    ],
    indirect=True,
)
def test_abstract_classes(global_ns):
    ns = global_ns.namespace("abstract_classes")
    abstract_i = ns.class_("abstract_i")
    assert abstract_i.is_abstract is True
    derived_abstract_i = ns.class_("derived_abstract_i")
    assert derived_abstract_i.is_abstract is True
    implementation = ns.class_("implementation")
    assert implementation.is_abstract is not True


@pytest.mark.parametrize(
    "global_ns",
    [
        "global_ns_fixture_all_at_once1",
        "global_ns_fixture_all_at_once2",
        "global_ns_fixture_file_by_file1",
        "global_ns_fixture_file_by_file2",
    ],
    indirect=True,
)
def test_byte_size(global_ns):
    mptrs = global_ns.class_("members_pointers_t")
    assert mptrs.byte_size != 0


@pytest.mark.parametrize(
    "global_ns",
    [
        "global_ns_fixture_all_at_once1",
        "global_ns_fixture_all_at_once2",
        "global_ns_fixture_file_by_file1",
        "global_ns_fixture_file_by_file2",
    ],
    indirect=True,
)
def test_byte_align(global_ns):
    mptrs = global_ns.class_("members_pointers_t")
    assert mptrs.byte_align != 0


@pytest.mark.parametrize(
    "global_ns",
    [
        "global_ns_fixture_all_at_once1",
        "global_ns_fixture_all_at_once2",
        "global_ns_fixture_file_by_file1",
        "global_ns_fixture_file_by_file2",
    ],
    indirect=True,
)
def test_byte_offset(global_ns):
    mptrs = global_ns.class_("members_pointers_t")
    assert mptrs.variable("xxx").byte_offset != 0



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

import pytest

from . import autoconfig

from pygccxml import parser
from pygccxml import declarations


TEST_FILES = [
    "type_traits.hpp",
]


@pytest.fixture
def decls():
    COMPILATION_MODE = parser.COMPILATION_MODE.ALL_AT_ONCE
    config = autoconfig.cxx_parsers_cfg.config.clone()
    decls = parser.parse(TEST_FILES, config, COMPILATION_MODE)
    return decls


def __test_type_category(decls, ns_name, controller):
    ns_control = declarations.find_declaration(
        decls,
        decl_type=declarations.namespace_t,
        name=ns_name)
    assert ns_control is not None
    ns_yes = declarations.find_declaration(
        ns_control,
        decl_type=declarations.namespace_t,
        name='yes')
    assert ns_yes is not None
    ns_no = declarations.find_declaration(
        ns_control,
        decl_type=declarations.namespace_t,
        name='no')
    assert ns_no is not None
    for decl in ns_yes.declarations:
        if isinstance(decl, declarations.variable_t):
            assert controller(decl.decl_type) is not None
        elif isinstance(decl, declarations.calldef_t) and \
                decl.name.startswith('test_'):
            continue
        else:
            assert controller(decl) is not None
    for decl in ns_no.declarations:
        if isinstance(decl, declarations.calldef_t) and \
                decl.name.startswith('test_'):
            continue
        val = controller(decl)
        # TODO: This looks bad and should be improved?
        # Why a boolean or None?
        assert val is False or val is None


def __test_type_transformation(decls, ns_name, transformer):
    ns_control = declarations.find_declaration(
        decls,
        decl_type=declarations.namespace_t,
        name=ns_name)
    assert ns_control is not None
    ns_before = declarations.find_declaration(
        ns_control,
        decl_type=declarations.namespace_t,
        name='before')
    assert ns_before is not None
    ns_after = declarations.find_declaration(
        ns_control,
        decl_type=declarations.namespace_t,
        name='after')
    assert ns_after is not None

    for tbefore in ns_before.declarations:
        tafter = declarations.find_declaration(
            ns_after,
            name=tbefore.name)
        assert tafter is not None
        transformed = transformer(tbefore)
        assert declarations.is_same(
                transformed,
                tafter) is True


def test_is_enum(decls):
    __test_type_category(decls, 'is_enum', declarations.is_enum)


def test_is_void(decls):
    __test_type_category(decls, 'is_void', declarations.is_void)


def test_is_bool(decls):
    __test_type_category(decls, 'is_bool', declarations.is_bool)


def test_is_integral(decls):
    __test_type_category(decls, 'is_integral', declarations.is_integral)


def test_is_pointer(decls):
    __test_type_category(decls, 'is_pointer', declarations.is_pointer)


def test_is_void_pointer(decls):
    __test_type_category(
        decls, 'is_void_pointer', declarations.is_void_pointer)


def test_is_const(decls):
    __test_type_category(decls, 'is_const', declarations.is_const)


def test_is_volatile(decls):
    __test_type_category(decls, 'is_volatile', declarations.is_volatile)


def test_is_reference(decls):
    __test_type_category(decls, 'is_reference', declarations.is_reference)


def test_is_floating_point(decls):
    __test_type_category(
        decls,
        'is_floating_point',
        declarations.is_floating_point)


def test_is_array(decls):
    __test_type_category(decls, 'is_array', declarations.is_array)


def test_is_fundamental(decls):
    __test_type_category(
        decls,
        'is_fundamental',
        declarations.is_fundamental)


def test_is_noncopyable(decls):
    __test_type_category(
        decls,
        'is_noncopyable',
        declarations.is_noncopyable)


def test_is_std_ostream(decls):
    __test_type_category(
        decls,
        'is_std_ostream',
        declarations.is_std_ostream)


def test_is_std_wostream(decls):
    __test_type_category(
        decls,
        'is_std_wostream',
        declarations.is_std_wostream)


def test_is_calldef_pointer(decls):
    __test_type_category(
        decls,
        'is_calldef_pointer',
        declarations.is_calldef_pointer)


def test_has_trivial_constructor(decls):
    __test_type_category(
        decls,
        'has_trivial_constructor',
        declarations.has_trivial_constructor)


def test_has_public_constructor(decls):
    __test_type_category(
        decls,
        'has_public_constructor',
        declarations.has_public_constructor)


def test_has_public_destructor(decls):
    __test_type_category(
        decls,
        'has_public_destructor',
        declarations.has_public_destructor)


def test_has_any_non_copyconstructor(decls):
    __test_type_category(
        decls,
        'has_any_non_copyconstructor',
        declarations.has_any_non_copyconstructor)


def test_has_copy_constructor(decls):
    __test_type_category(
        decls,
        'has_copy_constructor',
        declarations.has_copy_constructor)


def test_is_base_and_derived(decls):
    ns = declarations.find_declaration(
        decls,
        decl_type=declarations.namespace_t,
        name='is_base_and_derived')
    assert ns is not None
    base = declarations.find_declaration(
        ns.declarations,
        decl_type=declarations.class_t,
        name='base')
    derived = declarations.find_declaration(
        ns.declarations,
        decl_type=declarations.class_t,
        name='derived')
    assert base is not None
    assert derived is not None
    assert declarations.is_base_and_derived(base, derived) is True
    assert declarations.is_base_and_derived(base, (derived, derived)) is True

    unrelated1 = declarations.find_declaration(
        ns.declarations,
        decl_type=declarations.class_t,
        name='unrelated1')

    unrelated2 = declarations.find_declaration(
        ns.declarations,
        decl_type=declarations.class_t,
        name='unrelated2')
    assert base is not None
    assert derived is not None
    assert declarations.is_base_and_derived(unrelated1, unrelated2) is False


def test_is_same():
    assert declarations.is_same(
        declarations.int_t,
        declarations.int_t) is True
    assert declarations.is_same(
        declarations.int_t,
        declarations.float_t) is False


def test_remove_const(decls):
    __test_type_transformation(
        decls,
        'remove_const',
        declarations.remove_const)


def test_remove_reference(decls):
    __test_type_transformation(
        decls,
        'remove_reference',
        declarations.remove_reference)


def test_remove_volatile(decls):
    __test_type_transformation(
        decls,
        'remove_volatile',
        declarations.remove_volatile)


def test_remove_cv(decls):
    __test_type_transformation(
        decls, 'remove_cv', declarations.remove_cv
    )


def test_remove_pointer(decls):
    __test_type_transformation(
        decls,
        'remove_pointer',
        declarations.remove_pointer)


def test_is_unary_binary_operator(decls):
    operator_not = declarations.find_declaration(
        decls,
        decl_type=declarations.operator_t,
        fullname='::is_unary_operator::dummy::operator!')
    assert operator_not is not None
    assert declarations.is_unary_operator(operator_not) is True
    assert declarations.is_binary_operator(operator_not) is False

    operator_class_p = declarations.find_declaration(
        decls,
        decl_type=declarations.operator_t,
        fullname='::is_unary_operator::dummy::operator+')
    assert operator_class_p is not None
    assert declarations.is_unary_operator(operator_class_p) is False
    assert declarations.is_binary_operator(operator_class_p) is True

    operator_class_pp = declarations.find_declaration(
        decls,
        decl_type=declarations.operator_t,
        fullname='::is_unary_operator::dummy::operator++')
    assert operator_class_pp is not None
    assert declarations.is_unary_operator(operator_class_pp) is True
    assert declarations.is_binary_operator(operator_class_pp) is False

    operator_pp = declarations.find_declaration(
        decls,
        decl_type=declarations.operator_t,
        fullname='::is_unary_operator::operator++')
    assert operator_pp is not None
    assert declarations.is_unary_operator(operator_pp) is True
    assert declarations.is_binary_operator(operator_pp) is False

    operator_mm = declarations.find_declaration(
        decls,
        decl_type=declarations.operator_t,
        fullname='::is_unary_operator::operator*')
    assert operator_mm is not None
    assert declarations.is_unary_operator(operator_mm) is False
    assert declarations.is_binary_operator(operator_mm) is True

    operator_pe = declarations.find_declaration(
        decls,
        decl_type=declarations.operator_t,
        fullname='::is_unary_operator::operator+=')
    assert operator_pe is not None
    assert declarations.is_unary_operator(operator_pe) is False
    assert declarations.is_binary_operator(operator_pe) is True


def __is_convertible_impl(decl):
    defs = decl.bases[0].related_class.declarations
    source_type = declarations.find_declaration(defs, name='source_type')
    target_type = declarations.find_declaration(defs, name='target_type')
    expected_type = declarations.find_declaration(
        defs,
        name='expected',
        decl_type=declarations.enumeration_t)
    expected_value = bool(expected_type.get_name2value_dict()['value'])
    assert expected_value == declarations.is_convertible(
        source_type,
        target_type
    )


def test_is_convertible(decls):
    ns_is_convertible = declarations.find_declaration(
        decls,
        decl_type=declarations.namespace_t,
        name="is_convertible")

    assert ns_is_convertible is not None
    for tester in [
        decl for decl in ns_is_convertible.declarations if
            decl.name.startswith('x')]:
        __is_convertible_impl(tester)


def test_missing_decls():
    config = autoconfig.cxx_parsers_cfg.config
    code = "struct const_item{ const int values[10]; };"
    global_ns = parser.parse_string(code, config)[0]
    ci = global_ns.class_('const_item')
    assert len(ci.declarations) == 3


def test_get_declaration():
    code = """
        namespace A{
        struct B{
            int c;
        };

        template <class T>
        struct C: public T{
            int d;
        };

        template <class T>
        struct D{
            int dD;
        };

        typedef C<B> easy;
        typedef D<easy> Deasy;

        inline void instantiate(){
            int val = sizeof(easy);
        }

        }
    """

    global_ns = parser.parse_string(
        code,
        autoconfig.cxx_parsers_cfg.config)
    global_ns = declarations.get_global_namespace(global_ns)
    easy = global_ns.typedef('easy')
    declarations.class_traits.get_declaration(easy)
    deasy = global_ns.typedef('Deasy')
    d_a = declarations.class_traits.get_declaration(deasy)
    assert isinstance(d_a, declarations.class_types)



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

import pytest

from . import autoconfig

from pygccxml import parser
from pygccxml import declarations

TEST_FILES = ['test_function_pointer.hpp']


@pytest.fixture
def global_ns():
    COMPILATION_MODE = parser.COMPILATION_MODE.ALL_AT_ONCE
    config = autoconfig.cxx_parsers_cfg.config.clone()
    decls = parser.parse(TEST_FILES, config, COMPILATION_MODE)
    global_ns = declarations.get_global_namespace(decls)
    global_ns.init_optimizer()
    return global_ns


def test_function_pointer(global_ns):
    """
    Test working with pointers and function pointers.

    """

    # Test on a function pointer
    criteria = declarations.variable_matcher(name="func1")
    variables = declarations.matcher.find(criteria, global_ns)

    assert variables[0].name == "func1"
    assert isinstance(variables[0].decl_type, declarations.pointer_t) is True
    assert str(variables[0].decl_type) == "void (*)( int,double )"
    assert declarations.is_calldef_pointer(variables[0].decl_type) is True
    assert isinstance(
        declarations.remove_pointer(variables[0].decl_type),
        declarations.free_function_type_t
        ) is True

    # Get the function (free_function_type_t) and test the return and
    # argument types
    function = variables[0].decl_type.base
    assert isinstance(function.return_type, declarations.void_t)
    assert isinstance(function.arguments_types[0], declarations.int_t)
    assert isinstance(function.arguments_types[1], declarations.double_t)

    # Test on a normal pointer
    criteria = declarations.variable_matcher(name="myPointer")
    variables = declarations.matcher.find(criteria, global_ns)

    assert variables[0].name == "myPointer"
    assert isinstance(variables[0].decl_type, declarations.pointer_t)
    assert declarations.is_calldef_pointer(variables[0].decl_type) is False
    assert isinstance(
        declarations.remove_pointer(variables[0].decl_type),
        declarations.volatile_t
        ) is True

    # Test on function pointer in struct (x8)
    for d in global_ns.declarations:
        if d.name == "x8":
            assert isinstance(d.decl_type, declarations.pointer_t) is True
            assert declarations.is_calldef_pointer(d.decl_type)
            assert isinstance(
                declarations.remove_pointer(d.decl_type),
                declarations.member_function_type_t
                ) is True
            assert str(declarations.remove_pointer(d.decl_type)) == \
                "void ( ::some_struct_t::* )(  )"



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

import pytest

from . import autoconfig

from pygccxml import parser
from pygccxml import declarations

TEST_FILES = [
    "declarations_enums.hpp",
    "declarations_variables.hpp",
    "declarations_calldef.hpp",
]


@pytest.fixture
def global_ns_fixture_all_at_once():
    COMPILATION_MODE = parser.COMPILATION_MODE.ALL_AT_ONCE
    config = autoconfig.cxx_parsers_cfg.config.clone()
    decls = parser.parse(TEST_FILES, config, COMPILATION_MODE)
    global_ns = declarations.get_global_namespace(decls)
    return global_ns


@pytest.fixture
def global_ns_fixture_file_by_file():
    COMPILATION_MODE = parser.COMPILATION_MODE.FILE_BY_FILE
    config = autoconfig.cxx_parsers_cfg.config.clone()
    decls = parser.parse(TEST_FILES, config, COMPILATION_MODE)
    global_ns = declarations.get_global_namespace(decls)
    return global_ns


@pytest.fixture
def global_ns(request):
    return request.getfixturevalue(request.param)


@pytest.mark.parametrize(
    "global_ns",
    [
        "global_ns_fixture_all_at_once",
        "global_ns_fixture_file_by_file",
    ],
    indirect=True,
)
def test_enumeration_t(global_ns):
    enum = global_ns.enumeration("ENumbers")
    expected_values = list(
        zip(
            ["e%d" % index for index in range(10)],
            [index for index in range(10)]
        )
    )
    assert expected_values == enum.values


def test_namespace():
    pass  # tested in core_tester


def test_types():
    pass  # tested in core_tester


@pytest.mark.parametrize(
    "global_ns",
    [
        "global_ns_fixture_all_at_once",
        "global_ns_fixture_file_by_file",
    ],
    indirect=True,
)
def test_variables(global_ns, helpers):
    global_ns.namespace("variables")
    initialized = global_ns.variable(name="initialized")

    expected_value = "10122004"
    assert initialized.value == expected_value
    helpers._test_type_composition(
        initialized.decl_type,
        declarations.const_t,
        declarations.long_unsigned_int_t
    )

    m_mutable = global_ns.variable(name="m_mutable")
    assert m_mutable.type_qualifiers.has_static is False
    assert m_mutable.type_qualifiers.has_mutable is True

    # External static variable
    extern_var = global_ns.variable(name="extern_var")
    assert extern_var.type_qualifiers.has_extern is True
    assert extern_var.type_qualifiers.has_static is False
    assert extern_var.type_qualifiers.has_mutable is False

    # Static variable
    static_var = global_ns.variable(name="static_var")
    assert static_var.type_qualifiers.has_static is True
    assert static_var.type_qualifiers.has_extern is False
    assert static_var.type_qualifiers.has_mutable is False

    ssv_static_var = global_ns.variable(name="ssv_static_var")
    assert ssv_static_var.type_qualifiers.has_static is True
    assert ssv_static_var.type_qualifiers.has_extern is False
    assert ssv_static_var.type_qualifiers.has_mutable is False

    ssv_static_var_value = global_ns.variable(name="ssv_static_var_value")
    assert ssv_static_var_value.type_qualifiers.has_static is True
    assert ssv_static_var_value.type_qualifiers.has_extern is False
    assert ssv_static_var_value.type_qualifiers.has_mutable is False


@pytest.mark.parametrize(
    "global_ns",
    [
        "global_ns_fixture_all_at_once",
        "global_ns_fixture_file_by_file",
    ],
    indirect=True,
)
def test_calldef_free_functions(global_ns, helpers):
    ns = global_ns.namespace("calldef")

    no_return_no_args = ns.free_function("no_return_no_args")

    helpers._test_calldef_return_type(no_return_no_args, declarations.void_t)
    assert no_return_no_args.has_extern is False

    # Static_call is explicetely defined as extern, this works with gccxml
    # and castxml.
    static_call = ns.free_function("static_call")
    assert static_call is not None

    return_no_args = ns.free_function("return_no_args")
    helpers._test_calldef_return_type(return_no_args, declarations.int_t)
    # from now there is no need to check return type.
    no_return_1_arg = ns.free_function(name="no_return_1_arg")
    assert no_return_1_arg is not None
    assert no_return_1_arg.arguments[0].name in ["arg", "arg0"]
    helpers._test_calldef_args(
        no_return_1_arg,
        [
            declarations.argument_t(
                name=no_return_1_arg.arguments[0].name,
                decl_type=declarations.int_t()
            )
        ],
    )

    return_default_args = ns.free_function("return_default_args")
    assert return_default_args.arguments[0].name in ["arg", "arg0"]
    assert return_default_args.arguments[1].name in ["arg1", "flag"]
    helpers._test_calldef_args(
        return_default_args,
        [
            declarations.argument_t(
                name=return_default_args.arguments[0].name,
                decl_type=declarations.int_t(),
                default_value="1",
            ),
            declarations.argument_t(
                name=return_default_args.arguments[1].name,
                decl_type=declarations.bool_t(),
                default_value="false",
            ),
        ],
    )
    helpers._test_calldef_exceptions(global_ns, return_default_args, [])


@pytest.mark.parametrize(
    "global_ns",
    [
        "global_ns_fixture_all_at_once",
        "global_ns_fixture_file_by_file",
    ],
    indirect=True,
)
def test_calldef_member_functions(global_ns, helpers):
    struct_calldefs = global_ns.class_("calldefs_t")

    member_inline_call = struct_calldefs.member_function("member_inline_call")
    helpers._test_calldef_args(
        member_inline_call,
        [declarations.argument_t(name="i", decl_type=declarations.int_t())],
    )

    member_const_call = struct_calldefs.member_function("member_const_call")
    assert member_const_call.has_const
    assert member_const_call.virtuality == \
        declarations.VIRTUALITY_TYPES.NOT_VIRTUAL

    member_virtual_call = struct_calldefs.member_function(
        name="member_virtual_call"
        )
    assert member_virtual_call.virtuality == \
        declarations.VIRTUALITY_TYPES.VIRTUAL

    member_pure_virtual_call = struct_calldefs.member_function(
        "member_pure_virtual_call"
    )
    assert (
        member_pure_virtual_call.virtuality
        == declarations.VIRTUALITY_TYPES.PURE_VIRTUAL
    )

    static_call = struct_calldefs.member_function("static_call")
    assert static_call.has_static is True


@pytest.mark.parametrize(
    "global_ns",
    [
        "global_ns_fixture_all_at_once",
        "global_ns_fixture_file_by_file",
    ],
    indirect=True,
)
def test_constructors_destructors(global_ns, helpers):
    struct_calldefs = global_ns.class_("calldefs_t")

    destructor = struct_calldefs.calldef("~calldefs_t")
    helpers._test_calldef_args(destructor, [])
    helpers._test_calldef_return_type(destructor, None.__class__)

    # well, now we have a few functions ( constructors ) with the same
    # name, there is no easy way to find the desired one. Well in my case
    # I have only 4 constructors
    # 1. from char
    # 2. from (int,double)
    # 3. default
    # 4. copy constructor
    constructor_found = struct_calldefs.constructors("calldefs_t")
    assert len(constructor_found) == 5
    assert (
        len(
            [
                constructor
                for constructor in constructor_found
                if declarations.is_copy_constructor(constructor)
            ]
        )
        == 1
    )
    # there is nothing to check about constructors - I know the
    # implementation of parser.
    # In this case it doesn't different from any other function

    c = struct_calldefs.constructor("calldefs_t", arg_types=["char"])
    assert c.explicit is True

    arg_type = declarations.declarated_t(global_ns.class_("some_exception_t"))
    c = struct_calldefs.constructor("calldefs_t", arg_types=[arg_type])
    assert c.explicit is False


@pytest.mark.parametrize(
    "global_ns",
    [
        "global_ns_fixture_all_at_once",
        "global_ns_fixture_file_by_file",
    ],
    indirect=True,
)
def test_operator_symbol(global_ns):
    calldefs_operators = ["=", "=="]
    calldefs_cast_operators = ["char *", "double"]
    struct_calldefs = global_ns.class_("calldefs_t")
    assert struct_calldefs is not None
    for decl in struct_calldefs.declarations:
        if not isinstance(decl, declarations.operator_t):
            continue
        if not isinstance(decl, declarations.casting_operator_t):
            assert decl.symbol in calldefs_operators
        else:
            assert decl.return_type.decl_string in calldefs_cast_operators


@pytest.mark.parametrize(
    "global_ns",
    [
        "global_ns_fixture_all_at_once",
        "global_ns_fixture_file_by_file",
    ],
    indirect=True,
)
def test_ellipsis(global_ns):
    ns = global_ns.namespace("ellipsis_tester")
    do_smth = ns.member_function("do_smth")
    assert do_smth.has_ellipsis is True
    do_smth_else = ns.free_function("do_smth_else")
    assert do_smth_else.has_ellipsis is True



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

from . import namespace
from . import calldef
from . import cpptypes
from . import type_traits
from . import class_declaration


class impl_details(object):

    """implementation details"""
    @staticmethod
    def is_defined_in_xxx(xxx, cls):
        """
        Small helper method that checks whether the class `cls` is defined
        under `::xxx` namespace
        """
        if not cls.parent:
            return False

        if not isinstance(cls.parent, namespace.namespace_t):
            return False

        if xxx != cls.parent.name:
            return False

        xxx_ns = cls.parent
        if not xxx_ns.parent:
            return False

        if not isinstance(xxx_ns.parent, namespace.namespace_t):
            return False

        if xxx_ns.parent.name != '::':
            return False

        global_ns = xxx_ns.parent
        return None is global_ns.parent

    @staticmethod
    def find_value_type(global_ns, value_type_str):
        """implementation details"""
        if not value_type_str.startswith('::'):
            value_type_str = '::' + value_type_str
        found = global_ns.decls(
            name=value_type_str,
            function=lambda decl: not isinstance(decl, calldef.calldef_t),
            allow_empty=True)

        if len(found) == 1:
            return found[0]

        no_global_ns_value_type_str = value_type_str[2:]
        if no_global_ns_value_type_str in cpptypes.FUNDAMENTAL_TYPES:
            return cpptypes.FUNDAMENTAL_TYPES[no_global_ns_value_type_str]
        elif type_traits.is_std_string(value_type_str):
            string_ = global_ns.typedef('::std::string')
            return type_traits.remove_declarated(string_)
        elif type_traits.is_std_wstring(value_type_str):
            string_ = global_ns.typedef('::std::wstring')
            return type_traits.remove_declarated(string_)
        else:
            value_type_str = no_global_ns_value_type_str
            has_const = value_type_str.startswith('const ')
            if has_const:
                value_type_str = value_type_str[len('const '):]
            has_pointer = value_type_str.endswith('*')
            if has_pointer:
                value_type_str = value_type_str[:-1]
            found = None
            if has_const or has_pointer:
                found = impl_details.find_value_type(
                    global_ns,
                    value_type_str)
            if not found:
                return None
            else:
                if isinstance(found, class_declaration.class_types):
                    return cpptypes.declarated_t(found)
                if has_const:
                    return cpptypes.const_t(found)
                if has_pointer:
                    return cpptypes.pointer_t(found)

        return None



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

import pytest

from . import autoconfig

from pygccxml import parser
from pygccxml import declarations
from pygccxml.declarations import type_traits


TEST_FILES = [
    "unnamed_classes.hpp",
]


@pytest.fixture
def global_ns():
    COMPILATION_MODE = parser.COMPILATION_MODE.ALL_AT_ONCE
    config = autoconfig.cxx_parsers_cfg.config.clone()
    decls = parser.parse(TEST_FILES, config, COMPILATION_MODE)
    global_ns = declarations.get_global_namespace(decls)
    global_ns.init_optimizer()
    return global_ns


def validate_bitfields(parent, bitfields):
    for key in bitfields:
        var = parent.variable(key)
        assert var.bits == bitfields[key]


def do_union_test(global_ns, union_name, bitfields):
    s2 = global_ns.class_('S2')
    assert declarations.is_union(s2) is False
    assert declarations.is_struct(s2) is True
    assert s2.parent.name == 'S1'
    assert declarations.is_union(s2.parent) is False

    union = s2.variable(union_name)
    assert declarations.is_union(union.decl_type) is True
    assert declarations.is_struct(union.decl_type) is False

    union_type = type_traits.remove_declarated(union.decl_type)
    validate_bitfields(union_type, bitfields)
    assert union_type.variable('raw') is not None


def test_union_Flags(global_ns):
    flags_bitfields = {
        'hasItemIdList': 1,
        'pointsToFileOrDir': 1,
        'hasDescription': 1,
        'hasRelativePath': 1,
        'hasWorkingDir': 1,
        'hasCmdLineArgs': 1,
        'hasCustomIcon': 1,
        'useWorkingDir': 1,
        'unused': 24,
    }
    do_union_test(global_ns, 'flags', flags_bitfields)


def test_unnamed_unions(global_ns):
    fileattribs_bitfields = {
        'isReadOnly': 1,
        'isHidden': 1,
        'isSystem': 1,
        'isVolumeLabel': 1,
        'isDir': 1,
        'isModified': 1,
        'isEncrypted': 1,
        'isNormal': 1,
        'isTemporary': 1,
        'isSparse': 1,
        'hasReparsePoint': 1,
        'isCompressed': 1,
        'isOffline': 1,
        'unused': 19,
    }
    do_union_test(global_ns, 'fileattribs', fileattribs_bitfields)


def test_anonymous_unions(global_ns):
    s3 = global_ns.class_('S3')
    assert s3.parent.name == 'S1'

    s3_vars = ['anon_mem_c', 'anon_mem_i', 's3_mem', 's2']
    for var in s3_vars:
        assert declarations.is_union(s3.variable(var).decl_type) is False



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

from pygccxml import utils
from pygccxml import declarations
from pygccxml import parser

import os
import sys
import warnings
warnings.simplefilter("error", Warning)
# Find out the file location within the sources tree
this_module_dir_path = os.path.abspath(
    os.path.dirname(sys.modules[__name__].__file__))

# Find out the c++ parser
generator_path, generator_name = utils.find_xml_generator()

# Configure the xml generator
xml_generator_config = parser.xml_generator_configuration_t(
    xml_generator_path=generator_path,
    xml_generator=generator_name)

# The c++ file we want to parse
filename = "example.hpp"
filename = this_module_dir_path + "/" + filename

file_config = parser.file_configuration_t(
    data=filename,
    content_type=parser.CONTENT_TYPE.CACHED_SOURCE_FILE)

project_reader = parser.project_reader_t(xml_generator_config)
decls = project_reader.read_files(
    [file_config],
    compilation_mode=parser.COMPILATION_MODE.FILE_BY_FILE)

global_namespace = declarations.get_global_namespace(decls)

value = global_namespace.namespace("ns")
print("My name is: " + value.name)



# Copyright 2014-2016 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

from . import autoconfig

from pygccxml import parser
from pygccxml import declarations

TEST_FILES = ["test_non_copyable_recursive.hpp"]


def test_infinite_recursion_base_classes():
    """
    Test find_noncopyable_vars

    See #71

    find_noncopyable_vars was throwing:
    RuntimeError: maximum recursion depth exceeded while
    calling a Python object
    """
    config = autoconfig.cxx_parsers_cfg.config.clone()
    decls = parser.parse(TEST_FILES, config)
    global_ns = declarations.get_global_namespace(decls)

    # Description of the problem (before the fix):
    # find_noncopyable_vars (on Child class) looks up the variables,
    # and finds aBasePtr2 (a pointer to the Base2 class).
    # Then it looks recursively at the base classes of Base2, and finds
    # Base1. Then, it looks up the variables from Base, to check if Base1
    # is non copyable. It finds another aBasePtr2 variable, which leads to
    # a new check of Base2; this recurses infinitely.
    test_ns = global_ns.namespace('Test1')
    cls = test_ns.class_('Child')
    declarations.type_traits_classes.find_noncopyable_vars(cls)
    assert declarations.type_traits_classes.is_noncopyable(cls) is True


def test_infinite_recursion_sstream():
    """
    Test find_noncopyable_vars

    See #71

    find_noncopyable_vars was throwing:
    RuntimeError: maximum recursion depth exceeded while
    calling a Python object
    """
    config = autoconfig.cxx_parsers_cfg.config.clone()
    decls = parser.parse(TEST_FILES, config)
    global_ns = declarations.get_global_namespace(decls)

    # Real life example of the bug. This leads to a similar error,
    # but the situation is more complex as there are multiple
    # classes that are related the one to the others
    # (basic_istream, basic_ios, ios_base, ...)
    test_ns = global_ns.namespace('Test2')
    cls = test_ns.class_('FileStreamDataStream')
    declarations.type_traits_classes.find_noncopyable_vars(cls)
    assert declarations.type_traits_classes.is_noncopyable(cls) is False



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

import os
import shutil

import pytest

from . import autoconfig

from pygccxml import parser

TEST_FILES = [
    "typedefs1.hpp"
]

CACHE_DIR = os.path.join(autoconfig.data_directory, "directory_cache_test")


def set_up():
    if os.path.isdir(CACHE_DIR):
        shutil.rmtree(CACHE_DIR)
    if os.path.isfile(CACHE_DIR):
        os.remove(CACHE_DIR)


def test_directory_cache_without_compression():
    """
    Test the directory cache without compression

    """
    config = autoconfig.cxx_parsers_cfg.config.clone()
    set_up()
    # Test with compression OFF
    cache = parser.directory_cache_t(directory=CACHE_DIR)
    # Generate a cache on first read
    parser.parse(TEST_FILES, config, cache=cache)
    # Read from the cache the second time
    parser.parse(TEST_FILES, config, cache=cache)


def test_directory_cache_with_compression():
    """
    Test the directory cache wit compression

    """
    config = autoconfig.cxx_parsers_cfg.config.clone()
    set_up()
    # Test with compression ON
    cache = parser.directory_cache_t(
        directory=CACHE_DIR, compression=True)
    # Generate a cache on first read
    parser.parse(TEST_FILES, config, cache=cache)
    # Read from the cache the second time
    parser.parse(TEST_FILES, config, cache=cache)


def test_directory_cache_twice():
    """
    Setup two caches in a row.

    The second run will reload the same cache directory.
    """
    config = autoconfig.cxx_parsers_cfg.config.clone()
    set_up()
    cache = parser.directory_cache_t(directory=CACHE_DIR)
    parser.parse(TEST_FILES, config, cache=cache)
    cache = parser.directory_cache_t(directory=CACHE_DIR)
    parser.parse(TEST_FILES, config, cache=cache)


def test_directory_existing_dir():
    """
    Setup a cache when there is already a file at the cache's location.
    """
    set_up()
    open(CACHE_DIR, "a").close()
    with pytest.raises(ValueError):
        parser.directory_cache_t(directory=CACHE_DIR)



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

import pytest

from . import autoconfig

from pygccxml import parser
from pygccxml import declarations

TEST_FILES = [
    "better_templates_matcher_tester.hpp",
]


@pytest.fixture
def global_ns():
    COMPILATION_MODE = parser.COMPILATION_MODE.ALL_AT_ONCE
    INIT_OPTIMIZER = True
    config = autoconfig.cxx_parsers_cfg.config.clone()
    decls = parser.parse(TEST_FILES, config, COMPILATION_MODE)
    global_ns = declarations.get_global_namespace(decls)
    if INIT_OPTIMIZER:
        global_ns.init_optimizer()
    return global_ns


def test_better_templates_matcher(global_ns):
    classes = [
        "::Ogre::PlaneBoundedVolume",
        "::Ogre::Plane",
        "::Ogre::Singleton<Ogre::PCZoneFactoryManager>",
        "::Ogre::PCZoneFactoryManager",
        ]
    for i in classes:
        global_ns.class_(i)



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

from . import declaration
from . import typedef
from . import cpptypes
from . import type_traits
from . import class_declaration
from . import variable
from . import calldef
from . import namespace


def get_dependencies_from_decl(decl, recursive=True):
    """
    Returns the list of all types and declarations the declaration depends on.

    """
    result = []
    if isinstance(decl, typedef.typedef_t) or \
            isinstance(decl, variable.variable_t):
        return [dependency_info_t(decl, decl.decl_type)]
    if isinstance(decl, namespace.namespace_t):
        if recursive:
            for d in decl.declarations:
                result.extend(get_dependencies_from_decl(d))
        return result
    if isinstance(decl, calldef.calldef_t):
        if decl.return_type:
            result.append(
                dependency_info_t(decl, decl.return_type, hint="return type"))
        for arg in decl.arguments:
            result.append(dependency_info_t(decl, arg.decl_type))
        for exc in decl.exceptions:
            result.append(dependency_info_t(decl, exc, hint="exception"))
        return result
    if isinstance(decl, class_declaration.class_t):
        for base in decl.bases:
            result.append(
                dependency_info_t(
                    decl,
                    base.related_class,
                    base.access_type,
                    "base class"))
        if recursive:
            for access_type in class_declaration.ACCESS_TYPES.ALL:
                result.extend(
                    __find_out_member_dependencies(
                        decl.get_members(access_type), access_type))
        return result
    return result


def __find_out_member_dependencies(members, access_type):
    answer = []
    for mem in members:
        answer.extend(get_dependencies_from_decl(mem, recursive=True))
    member_ids = set([id(m) for m in members])
    for dependency in answer:
        if id(dependency.declaration) in member_ids:
            dependency.access_type = access_type
    return answer


class dependency_info_t(object):

    def __init__(self, decl, depend_on_it, access_type=None, hint=None):
        object.__init__(self)

        assert isinstance(
            depend_on_it, (class_declaration.class_t, cpptypes.type_t))
        self._decl = decl
        self._depend_on_it = depend_on_it
        self._access_type = access_type
        self._hint = hint

    @property
    def declaration(self):
        return self._decl

    @property
    def depend_on_it(self):
        return self._depend_on_it

    @property
    def access_type(self):
        return self._access_type

    @access_type.setter
    def access_type(self, access_type):
        self._access_type = access_type

    def __str__(self):
        return 'declaration "%s" depends( %s ) on "%s" ' \
               % (self.declaration, self.access_type, self.depend_on_it)

    @property
    def hint(self):
        """The declaration, that report dependency can put some additional
        inforamtion about dependency. It can be used later"""
        return self._hint

    def find_out_depend_on_it_declarations(self):
        """If declaration depends on other declaration and not on some type
        this function will return reference to it. Otherwise None will be
        returned
        """
        return impl_details.dig_declarations(self.depend_on_it)

    @staticmethod
    def i_depend_on_them(decl):
        """Returns set of declarations. every item in the returned set,
        depends on a declaration from the input"""

        to_be_included = set()
        for dependency_info in get_dependencies_from_decl(decl):
            for ddecl in dependency_info.find_out_depend_on_it_declarations():
                if ddecl:
                    to_be_included.add(ddecl)

        if isinstance(decl.parent, class_declaration.class_t):
            to_be_included.add(decl.parent)
        return to_be_included

    @staticmethod
    def we_depend_on_them(decls):
        """Returns set of declarations. every item in the returned set,
        depends on a declaration from the input"""
        to_be_included = set()
        for decl in decls:
            to_be_included.update(dependency_info_t.i_depend_on_them(decl))
        return to_be_included


class impl_details(object):

    @staticmethod
    def dig_declarations(depend_on_it):

        if isinstance(depend_on_it, declaration.declaration_t):
            return [depend_on_it]
        base_type = type_traits.base_type(
            type_traits.remove_alias(depend_on_it))
        if isinstance(base_type, cpptypes.declarated_t):
            return [base_type.declaration]
        elif isinstance(base_type, cpptypes.calldef_type_t):
            result = []
            result.extend(impl_details.dig_declarations(base_type.return_type))
            for argtype in base_type.arguments_types:
                result.extend(impl_details.dig_declarations(argtype))
            if isinstance(base_type, cpptypes.member_function_type_t):
                result.extend(
                    impl_details.dig_declarations(
                        base_type.class_inst))
            return result
        return []



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

"""Parser sub-package.
"""

from .config import xml_generator_configuration_t
from .config import load_xml_generator_configuration

from .project_reader import COMPILATION_MODE
from .project_reader import project_reader_t
from .project_reader import file_configuration_t
from .project_reader import create_text_fc
from .project_reader import create_source_fc
from .project_reader import create_gccxml_fc
from .project_reader import create_cached_source_fc

from .source_reader import source_reader_t
from .declarations_cache import cache_base_t
from .declarations_cache import file_cache_t
from .declarations_cache import dummy_cache_t
from .directory_cache import directory_cache_t
# shortcut
CONTENT_TYPE = file_configuration_t.CONTENT_TYPE


def parse(
        files,
        config=None,
        compilation_mode=COMPILATION_MODE.FILE_BY_FILE,
        cache=None):
    """
    Parse header files.

    :param files: The header files that should be parsed
    :type files: list of str
    :param config: Configuration object or None
    :type config: :class:`parser.xml_generator_configuration_t`
    :param compilation_mode: Determines whether the files are parsed
                             individually or as one single chunk
    :type compilation_mode: :class:`parser.COMPILATION_MODE`
    :param cache: Declaration cache (None=no cache)
    :type cache: :class:`parser.cache_base_t` or str
    :rtype: list of :class:`declarations.declaration_t`
    """
    if not config:
        config = xml_generator_configuration_t()
    parser = project_reader_t(config=config, cache=cache)
    declarations = parser.read_files(files, compilation_mode)
    config.xml_generator_from_xml_file = parser.xml_generator_from_xml_file
    return declarations


def parse_string(content, config=None):
    if not config:
        config = xml_generator_configuration_t()
    parser = project_reader_t(config)
    declarations = parser.read_string(content)
    config.xml_generator_from_xml_file = parser.xml_generator_from_xml_file
    return declarations


def parse_xml_file(content, config=None):
    parser = source_reader_t(config)
    decls = parser.read_xml_file(content)
    config.xml_generator_from_xml_file = parser.xml_generator_from_xml_file
    return decls



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

import os
import sys

# Find out the file location within the sources tree
this_module_dir_path = os.path.abspath(
    os.path.dirname(sys.modules[__name__].__file__))
# Add pygccxml package to Python path
sys.path.append(os.path.join(this_module_dir_path, '..', '..'))


from pygccxml import parser  # nopep8
from pygccxml import declarations  # nopep8
from pygccxml import utils  # nopep8

# Find out the xml generator (gccxml or castxml)
generator_path, generator_name = utils.find_xml_generator()

# Configure the xml generator
config = parser.xml_generator_configuration_t(
    xml_generator_path=generator_path,
    xml_generator=generator_name,
    compiler="gcc")

# Parsing source file
decls = parser.parse([this_module_dir_path + '/example.hpp'], config)
global_ns = declarations.get_global_namespace(decls)

# Get object that describes unittests namespace
unittests = global_ns.namespace('unittests')

print('"unittests" declarations: \n')
declarations.print_declarations(unittests)

# Print all base and derived class names
for class_ in unittests.classes():
    print('class "%s" hierarchy information:' % class_.name)
    print('\tbase classes   : ', repr([
        base.related_class.name for base in class_.bases]))
    print('\tderived classes: ', repr([
        derive.related_class.name for derive in class_.derived]))
    print('\n')

# Pygccxml has very powerfull query api:

# Select multiple declarations
run_functions = unittests.member_functions('run')
print('the namespace contains %d "run" member functions' % len(run_functions))
print('they are: ')
for f in run_functions:
    print('\t' + declarations.full_name(f))

# Select single declaration - all next statements will return same object
# vector< unittests::test_case* >

# You can select the class using "full" name
test_container_1 = global_ns.class_('::unittests::test_suite')
# You can define your own "match" criteria
test_container_2 = global_ns.class_(lambda decl: 'suite' in decl.name)

is_same_object = test_container_1 is test_container_2
print(
    "Does all test_container_* refer to the same object? " +
    str(is_same_object))



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

import os
import pprint
import xml.sax
import xml.sax.handler
from .. import utils
from .. import declarations

# XML_NN - XML Node Name
# XML_AN - XML Attribute Name
XML_AN_ABSTRACT = "abstract"
XML_AN_ACCESS = "access"
XML_AN_ALIGN = "align"
XML_AN_ARTIFICIAL = "artificial"
XML_AN_ATTACHED = "attached"
XML_AN_ATTRIBUTES = "attributes"
XML_AN_BASE_TYPE = "basetype"
XML_AN_BASES = "bases"
XML_AN_BEGIN_COLUMN = "begin_column"
XML_AN_BEGIN_LINE = "begin_line"
XML_AN_BEGIN_OFFSET = "begin_offset"
XML_AN_BITS = "bits"
XML_AN_COMMENT = "comment"
XML_AN_CONST = "const"
XML_AN_CONTEXT = "context"
XML_AN_CVS_REVISION = "cvs_revision"
XML_AN_CASTXML_FORMAT = "format"
XML_AN_DEFAULT = "default"
XML_AN_DEPRECATION = "deprecation"
XML_AN_END_COLUMN = "end_column"
XML_AN_END_LINE = "end_line"
XML_AN_END_OFFSET = "end_offset"
XML_AN_EXPLICIT = "explicit"
XML_AN_EXTERN = "extern"
XML_AN_FILE = "file"
XML_AN_ID = "id"
XML_AN_INCOMPLETE = "incomplete"
XML_AN_INIT = "init"
XML_AN_INLINE = "inline"
XML_AN_LINE = "line"
XML_AN_MANGLED = "mangled"
XML_AN_MAX = "max"
XML_AN_MEMBERS = "members"
XML_AN_MUTABLE = "mutable"
XML_AN_NAME = "name"
XML_AN_OFFSET = "offset"
XML_AN_OVERRIDES = "overrides"
XML_AN_PURE_VIRTUAL = "pure_virtual"
XML_AN_RESTRICT = "restrict"
XML_AN_RETURNS = "returns"
XML_AN_SIZE = "size"
XML_AN_STATIC = "static"
XML_AN_THROW = "throw"
XML_AN_TYPE = "type"
XML_AN_ORIGINAL_TYPE = "original_type"
XML_AN_VIRTUAL = "virtual"
XML_AN_VOLATILE = "volatile"
XML_NN_ARGUMENT = "Argument"
XML_NN_ARRAY_TYPE = "ArrayType"
XML_NN_CASTING_OPERATOR = "Converter"
XML_NN_CLASS = "Class"
XML_NN_COMMENT = "Comment"
XML_NN_CONSTRUCTOR = "Constructor"
XML_NN_CV_QUALIFIED_TYPE = "CvQualifiedType"
XML_NN_DESTRUCTOR = "Destructor"
XML_NN_ELLIPSIS = "Ellipsis"
XML_NN_ENUMERATION = "Enumeration"
XML_NN_ENUMERATION_VALUE = "EnumValue"
XML_NN_FIELD = "Field"
XML_NN_FILE = "File"
XML_NN_FUNCTION = "Function"
XML_NN_FUNCTION_TYPE = "FunctionType"
XML_NN_FUNDAMENTAL_TYPE = "FundamentalType"
XML_NN_FREE_OPERATOR = "OperatorFunction"
XML_NN_GCC_XML = "GCC_XML"
XML_NN_CASTXML = "CastXML"
XML_NN_MEMBER_OPERATOR = "OperatorMethod"
XML_NN_METHOD = "Method"
XML_NN_METHOD_TYPE = "MethodType"
XML_NN_NAMESPACE = "Namespace"
XML_NN_OFFSET_TYPE = "OffsetType"
XML_NN_POINTER_TYPE = "PointerType"
XML_NN_REFERENCE_TYPE = "ReferenceType"
XML_NN_ELABORATED_TYPE = "ElaboratedType"
XML_NN_ROOT = "GCC_XML"
XML_NN_STRUCT = "Struct"
XML_NN_TYPEDEF = "Typedef"
XML_NN_UNION = "Union"
XML_NN_VARIABLE = "Variable"


class scanner_t(xml.sax.handler.ContentHandler):

    def __init__(self, xml_file, decl_factory, config, *args):
        xml.sax.handler.ContentHandler.__init__(self, *args)
        self.logger = utils.loggers.cxx_parser
        self.xml_file = xml_file
        self.config = config
        self.__readers = {
            XML_NN_FILE: self.__read_file,
            XML_NN_NAMESPACE: self.__read_namespace,
            XML_NN_ENUMERATION: self.__read_enumeration,
            XML_NN_ENUMERATION_VALUE: self.__read_enumeration_value,
            XML_NN_ARRAY_TYPE: self.__read_array_type,
            XML_NN_CV_QUALIFIED_TYPE: self.__read_cv_qualified_type,
            XML_NN_POINTER_TYPE: self.__read_pointer_type,
            XML_NN_REFERENCE_TYPE: self.__read_reference_type,
            XML_NN_ELABORATED_TYPE: self.__read_elaborated_type,
            XML_NN_FUNDAMENTAL_TYPE: self.__read_fundamental_type,
            XML_NN_ARGUMENT: self.__read_argument,
            XML_NN_FUNCTION_TYPE: self.__read_function_type,
            XML_NN_METHOD_TYPE: self.__read_method_type,
            XML_NN_OFFSET_TYPE: self.__read_offset_type,
            XML_NN_TYPEDEF: self.__read_typedef,
            XML_NN_VARIABLE: self.__read_variable,
            XML_NN_CLASS: self.__read_class,
            XML_NN_STRUCT: self.__read_struct,
            XML_NN_UNION: self.__read_union,
            XML_NN_FIELD: self.__read_field,
            XML_NN_CASTING_OPERATOR: self.__read_casting_operator,
            XML_NN_COMMENT: self.__read_comment,
            XML_NN_CONSTRUCTOR: self.__read_constructor,
            XML_NN_DESTRUCTOR: self.__read_destructor,
            XML_NN_FUNCTION: self.__read_function,
            XML_NN_FREE_OPERATOR: self.__read_free_operator,
            XML_NN_MEMBER_OPERATOR: self.__read_member_operator,
            XML_NN_METHOD: self.__read_method,
            XML_NN_GCC_XML: self.__read_version,
            XML_NN_CASTXML: self.__read_version,
            XML_NN_ELLIPSIS: self.__read_ellipsis}
        self.deep_declarations = [
            XML_NN_CASTING_OPERATOR,
            XML_NN_CONSTRUCTOR,
            XML_NN_DESTRUCTOR,
            XML_NN_ENUMERATION,
            XML_NN_FILE,
            XML_NN_COMMENT,
            XML_NN_FUNCTION,
            XML_NN_FREE_OPERATOR,
            XML_NN_MEMBER_OPERATOR,
            XML_NN_METHOD,
            XML_NN_FUNCTION_TYPE,
            XML_NN_METHOD_TYPE]

        assert isinstance(decl_factory, declarations.decl_factory_t)
        self.__decl_factory = decl_factory

        # mapping from id -> decl
        self.__declarations = {}
        # list of all read declarations
        self.__calldefs = []
        # list of enums I need later
        self.__enums = []
        # mapping of id -> comment decl
        self.__comments = {}
        # mapping from id -> type
        self.__types = {}
        # mapping from id -> file
        self.__files = {}
        # mapping between decl id -> access
        self.__access = {}
        # mapping between file decl_id -> text
        self.__files_text = {}
        # current object under construction
        self.__inst = None
        # mapping from id to members
        self.__members = {}

        self.__mangled_suffix = ' *INTERNAL* '
        self.__mangled_suffix_len = len(self.__mangled_suffix)

        # These fields are generated by clang, and have no location.
        # Just set an empty location for them.
        # bug #19: gp_offset, fp_offset, overflow_arg_area, reg_save_area
        # bug #32: isa, flags, str and length were added in llvm 3.9
        self.__name_attrs_to_skip = [
            "gp_offset",
            "fp_offset",
            "overflow_arg_area",
            "reg_save_area",
            "isa",
            "flags",
            "str",
            "length",
        ]

        # These fields are generated by GCC on AArch64, and have no location.
        self.__name_attrs_to_skip += [
            "__stack",
            "__gr_top",
            "__vr_top",
            "__gr_offs",
            "__vr_offs",
        ]

        # With CastXML and clang some __va_list_tag declarations are
        # present in the tree
        # With llvm 3.9 there is a __NSConstantString(_tag) in the tree
        self.__declarations_to_skip = [
            "__va_list_tag",
            "__va_list",
            "__NSConstantString_tag",
            "__NSConstantString"
        ]

        self.__xml_generator_from_xml_file = None

    @property
    def xml_generator_from_xml_file(self):
        """
        Configuration object containing information about the xml generator
        read from the xml file.

        Returns:
            utils.xml_generators: configuration object
        """
        return self.__xml_generator_from_xml_file

    def read(self):
        xml.sax.parse(self.xml_file, self)

    def _handle_comment(self, declaration):
        comm_decl = self.__comments.get(declaration.comment)
        if comm_decl:
            # First acquire file_name placeholder
            file_decl = comm_decl.location.file_name
            # If first encounter:
            # Find real name and read it into memory
            if file_decl not in self.__files_text:
                src_name = self.__files.get(file_decl)
                with open(src_name, "r") as file:
                    line_list = file.readlines()
                    self.__files_text[file_decl] = line_list
            comment_text = []
            # Capture file text for parsing
            file_text = self.__files_text.get(file_decl)
            # Use lines and columns to capture only comment text
            for indx in range(comm_decl.begin_line - 1, comm_decl.end_line):
                # Col data only different on the last line
                end_idx = -1
                strt_idx = comm_decl.begin_column - 1
                if indx == comm_decl.end_line - 1:
                    end_idx = comm_decl.end_column
                comm_line = file_text[indx]
                comm_line = comm_line[
                               strt_idx:end_idx
                ]
                # Remove newlines from end of string
                comm_line = comm_line.rstrip("\n")
                comment_text.append(comm_line)
            comm_decl.text = comment_text
            return comm_decl
        else:
            return declarations.comment.comment_t()

    def endDocument(self):
        # updating membership
        members_mapping = {}
        for gccxml_id, members in self.__members.items():
            decl = self.__declarations.get(gccxml_id)
            if not decl or not isinstance(decl, declarations.scopedef_t):
                continue
            members_mapping[id(decl)] = members
        self.__members = members_mapping
        for gccxml_id, decl in self.__declarations.items():
            if not isinstance(decl.comment, declarations.comment.comment_t):
                decl.comment = self._handle_comment(decl)
            if isinstance(decl, declarations.calldef_t) and decl.overrides:
                overrides_decl = self.__declarations.get(decl.overrides)
                decl.overrides = overrides_decl

    def declarations(self):
        return self.__declarations

    def calldefs(self):
        return self.__calldefs

    def enums(self):
        return self.__enums

    def types(self):
        return self.__types

    def files(self):
        return self.__files

    def access(self):
        return self.__access

    def members(self):
        return self.__members

    def startElement(self, name, attrs):
        try:
            if name not in self.__readers:
                return
            obj = self.__readers[name](attrs)
            if not obj:
                return  # it means that we worked on internals
                # for example EnumValue of function argument
            if name in self.deep_declarations:
                self.__inst = obj
            self.__read_access(attrs)
            element_id = attrs.get(XML_AN_ID)

            if isinstance(obj, declarations.declaration_t):

                if obj.name in self.__declarations_to_skip:
                    return
                self.__update_membership(attrs)
                self.__declarations[element_id] = obj
                if not isinstance(obj, declarations.namespace_t):
                    self.__read_location(obj, attrs, self.__name_attrs_to_skip)
                if isinstance(obj, declarations.class_t):
                    self.__read_bases(obj, attrs)
                self.__read_artificial(obj, attrs)
                self.__read_deprecation(obj, attrs)
                self.__read_mangled(obj, attrs)
                self.__read_attributes(obj, attrs)

            elif isinstance(obj, declarations.type_t):

                self.__types[element_id] = obj
                self.__read_byte_size(obj, attrs)
                self.__read_byte_align(obj, attrs)

            elif isinstance(obj, declarations.comment.comment_t):
                self.__comments[element_id] = obj
                self.__update_membership(attrs)
                self.__read_attributes(obj, attrs)

            elif isinstance(obj, str):

                self.__files[element_id] = os.path.normpath(obj)

            else:
                self.logger.warning(
                    'Unknown object type has been found.' +
                    ' Please report this bug to pygccxml development team.')
        except Exception as error:
            msg = (
                'error occured, while parsing element with name "%s" ' +
                'and attrs "%s".')
            msg = msg + os.linesep + 'Error: %s.' % str(error)
            self.logger.error(msg, name, pprint.pformat(list(attrs.keys())))
            raise

    def endElement(self, name):
        if name in self.deep_declarations:
            self.__inst = None

    @staticmethod
    def __read_location(decl, attrs, to_skip):
        if "name" in attrs and attrs["name"] in to_skip:
            decl.location = declarations.location_t('', -1)
        else:
            if XML_AN_BEGIN_LINE in attrs:
                line_number = attrs[XML_AN_BEGIN_LINE]
            else:
                line_number = attrs[XML_AN_LINE]
            decl.location = declarations.location_t(
                file_name=attrs[XML_AN_FILE],
                line=int(line_number))

    def __update_membership(self, attrs):
        parent = attrs.get(XML_AN_CONTEXT)
        if not parent:
            return
        if parent not in self.__members:
            self.__members[parent] = []
        self.__members[parent].append(attrs[XML_AN_ID])

    @staticmethod
    def __read_members(decl, attrs):
        decl.declarations = attrs.get(XML_AN_MEMBERS, "")

    @staticmethod
    def __read_bases(decl, attrs):
        decl.bases = attrs.get(XML_AN_BASES, "")

    @staticmethod
    def __read_artificial(decl, attrs):
        decl.is_artificial = attrs.get(XML_AN_ARTIFICIAL, False)

    def __read_mangled(self, decl, attrs):
        mangled = attrs.get(XML_AN_MANGLED)
        # the following patch is defined here for performance reasons
        if isinstance(mangled, bytes) and \
                mangled.endswith(self.__mangled_suffix):
            mangled = mangled[:self.__mangled_suffix_len]
        decl.mangled = mangled

    def __read_deprecation(self, decl, attrs):
        deprecation = attrs.get(XML_AN_DEPRECATION)
        # the following patch is defined here for performance reasons
        if isinstance(deprecation, str):
            decl.deprecation = deprecation

    def __read_attributes(self, decl, attrs):
        attribute = attrs.get(XML_AN_ATTRIBUTES)
        if attribute is not None and \
                self.config.compiler == "msvc" and \
                "f2" not in self.config.flags:
            if attribute == "__thiscall__":
                return
            if "__thiscall__" in attribute:
                attribute = attribute.replace("__thiscall__ ", "")
        decl.attributes = attribute

    def __read_access(self, attrs):
        self.__access[attrs[XML_AN_ID]] = \
            attrs.get(XML_AN_ACCESS, declarations.ACCESS_TYPES.PUBLIC)

    @staticmethod
    def __read_byte_size(decl, attrs):
        """Using duck typing to set the size instead of in constructor"""
        size = attrs.get(XML_AN_SIZE, 0)
        # Make sure the size is in bytes instead of bits
        decl.byte_size = int(size) / 8

    @staticmethod
    def __read_byte_offset(decl, attrs):
        """Using duck typing to set the offset instead of in constructor"""
        offset = attrs.get(XML_AN_OFFSET, 0)
        # Make sure the size is in bytes instead of bits
        decl.byte_offset = int(offset) / 8

    @staticmethod
    def __read_byte_align(decl, attrs):
        """Using duck typing to set the alignment"""
        align = attrs.get(XML_AN_ALIGN, 0)
        # Make sure the size is in bytes instead of bits
        decl.byte_align = int(align) / 8

    def __read_root(self, attrs):
        pass

    @staticmethod
    def __read_file(attrs):
        return attrs.get(XML_AN_NAME, '')

    def __read_namespace(self, attrs):
        ns_name = attrs.get(XML_AN_NAME, '')
        if '.' in ns_name:
            # if '.' in namespace then this is mangled namespace
            # -> in c++ namespace{...}
            # that is almost true: gcc mangale name using top file name.
            # almost all files has '.' in name
            ns_name = ''
        decl = self.__decl_factory.create_namespace(name=ns_name)
        if attrs.get(XML_AN_COMMENT) is not None:
            decl.comment = attrs.get(XML_AN_COMMENT)
        return decl

    def __read_enumeration(self, attrs):
        enum_name = attrs.get(XML_AN_NAME, '')
        if '$_' in enum_name or '._' in enum_name:
            # it means that this is unnamed enum. in c++ enum{ x };
            enum_name = ''
        decl = self.__decl_factory.create_enumeration(name=enum_name)
        if attrs.get(XML_AN_COMMENT) is not None:
            decl.comment = attrs.get(XML_AN_COMMENT)
        self.__read_byte_size(decl, attrs)
        self.__read_byte_align(decl, attrs)
        self.__enums.append(decl)
        return decl

    def __read_enumeration_value(self, attrs):
        name = attrs.get(XML_AN_NAME, '')
        num = int(attrs[XML_AN_INIT])
        self.__inst.append_value(name, num)

    @staticmethod
    def __guess_int_value(value_as_str):
        # returns instance of int or None
        # if gcc compiled the code, than it is correct!
        numeric_suffix_letters = 'UuLlFf'
        for s in numeric_suffix_letters:
            value_as_str = value_as_str.replace(s, '')
        try:
            return int(value_as_str)
        except ValueError:
            try:
                return int(value_as_str, 16)
            except ValueError:
                return None

    def __read_array_type(self, attrs):
        type_ = attrs[XML_AN_TYPE]
        size = self.__guess_int_value(attrs.get(XML_AN_MAX, ''))
        if size is None:
            size = declarations.array_t.SIZE_UNKNOWN
            # The following warning is pretty useless, as it cant say what the
            # problematic declaration is.
            # msg = 'unable to find out array size from expression
            # "%s"' % attrs[ XML_AN_MAX ]
            # warnings.warn( msg )
        return declarations.array_t(type_, size + 1)

    @staticmethod
    def __read_cv_qualified_type(attrs):
        if XML_AN_CONST in attrs and XML_AN_VOLATILE in attrs:
            return declarations.volatile_t(
                declarations.const_t(attrs[XML_AN_TYPE]))
        elif XML_AN_CONST in attrs:
            return declarations.const_t(attrs[XML_AN_TYPE])
        elif XML_AN_VOLATILE in attrs:
            return declarations.volatile_t(attrs[XML_AN_TYPE])
        elif XML_AN_RESTRICT in attrs:
            return declarations.restrict_t(attrs[XML_AN_TYPE])
        else:
            assert 0

    @staticmethod
    def __read_pointer_type(attrs):
        return declarations.pointer_t(attrs[XML_AN_TYPE])

    @staticmethod
    def __read_reference_type(attrs):
        return declarations.reference_t(attrs[XML_AN_TYPE])

    @staticmethod
    def __read_elaborated_type(attrs):
        return declarations.elaborated_t(attrs[XML_AN_TYPE])

    @staticmethod
    def __read_fundamental_type(attrs):
        try:
            return declarations.FUNDAMENTAL_TYPES[attrs.get(XML_AN_NAME, '')]
        except KeyError:
            return None
            # This code chokes on atomic_int_type in Boost 1.54
            # (and higher, probably).
            # It seems ok to just silently ignore this error.
            # raise RuntimeError((
            #    "pygccxml error: unable to find fundamental type with " +
            #    "name '%s'.") % attrs.get( XML_AN_NAME, '' ) )

    def __read_offset_type(self, attrs):
        base = attrs[XML_AN_BASE_TYPE]
        type_ = attrs[XML_AN_TYPE]
        return declarations.pointer_t(
            declarations.member_variable_type_t(
                class_inst=base,
                variable_type=type_))

    def __read_argument(self, attrs):
        if isinstance(self.__inst, declarations.calldef_type_t):
            self.__inst.arguments_types.append(attrs[XML_AN_TYPE])
        else:
            argument = declarations.argument_t()
            argument.name = attrs.get(
                XML_AN_NAME,
                'arg%d' % len(
                    self.__inst.arguments))
            argument.decl_type = attrs.get(
                XML_AN_ORIGINAL_TYPE,
                attrs.get(XML_AN_TYPE)
            )
            argument.default_value = attrs.get(XML_AN_DEFAULT)
            self.__read_attributes(argument, attrs)
            self.__inst.arguments.append(argument)

    def __read_ellipsis(self, attrs):
        if isinstance(self.__inst, declarations.calldef_type_t):
            self.__inst.arguments_types.append('...')
        else:
            argument = declarations.argument_t(decl_type='...')
            self.__inst.arguments.append(argument)

    def __read_calldef(self, calldef, attrs, is_declaration):
        # destructor for example doesn't have return type
        calldef.return_type = attrs.get(XML_AN_RETURNS)
        if is_declaration:
            self.__calldefs.append(calldef)
            calldef.name = attrs.get(XML_AN_NAME, '')
            calldef.has_extern = bool(attrs.get(XML_AN_EXTERN, False))
            calldef.has_inline = bool(attrs.get(XML_AN_INLINE, False))
            throw_stmt = attrs.get(XML_AN_THROW)
            if None is throw_stmt:
                calldef.does_throw = True
                calldef.exceptions = []
            elif throw_stmt == "":
                calldef.does_throw = False
                calldef.exceptions = []
            else:
                calldef.does_throw = True
                calldef.exceptions = throw_stmt.split()
        if attrs.get(XML_AN_COMMENT) is not None:
            calldef.comment = attrs.get(XML_AN_COMMENT)

    def __read_member_function(self, calldef, attrs, is_declaration):
        self.__read_calldef(calldef, attrs, is_declaration)
        calldef.has_const = bool(attrs.get(XML_AN_CONST, False))
        if is_declaration:
            calldef.has_static = bool(attrs.get(XML_AN_STATIC, False))
            if XML_AN_PURE_VIRTUAL in attrs:
                calldef.virtuality = declarations.VIRTUALITY_TYPES.PURE_VIRTUAL
            elif XML_AN_VIRTUAL in attrs:
                calldef.virtuality = declarations.VIRTUALITY_TYPES.VIRTUAL
            else:
                calldef.virtuality = declarations.VIRTUALITY_TYPES.NOT_VIRTUAL
            if XML_AN_OVERRIDES in attrs:
                calldef.overrides = attrs.get(XML_AN_OVERRIDES)
        else:
            calldef.class_inst = attrs[XML_AN_BASE_TYPE]

    def __read_function_type(self, attrs):
        answer = declarations.free_function_type_t()
        self.__read_calldef(answer, attrs, False)
        return answer

    def __read_method_type(self, attrs):
        answer = declarations.member_function_type_t()
        self.__read_member_function(answer, attrs, False)
        return answer

    def __read_typedef(self, attrs):
        return self.__decl_factory.create_typedef(
            name=attrs.get(
                XML_AN_NAME,
                ''),
            decl_type=attrs[XML_AN_TYPE])

    def __read_variable(self, attrs):
        type_qualifiers = declarations.type_qualifiers_t()
        type_qualifiers.has_mutable = bool(attrs.get(XML_AN_MUTABLE, False))
        type_qualifiers.has_static = bool(attrs.get(XML_AN_STATIC, False))
        type_qualifiers.has_extern = bool(attrs.get(XML_AN_EXTERN, False))
        bits = attrs.get(XML_AN_BITS)
        if bits:
            bits = int(bits)
        decl = self.__decl_factory.create_variable(
            name=attrs.get(
                XML_AN_NAME,
                ''),
            decl_type=attrs[XML_AN_TYPE],
            type_qualifiers=type_qualifiers,
            value=attrs.get(
                XML_AN_INIT),
            bits=bits)
        self.__read_byte_offset(decl, attrs)
        if attrs.get(XML_AN_COMMENT) is not None:
            decl.comment = attrs.get(XML_AN_COMMENT)
        return decl

    __read_field = __read_variable  # just a synonym

    def __read_class_impl(self, class_type, attrs):
        name = attrs.get(XML_AN_NAME, '')
        if '$' in name or '.' in name:
            name = ''
        if "<" in name and " >" in name:
            # Name with template. In some rare cases there
            # is a space before > (and only there), so remove
            # it to be consistent with the other names that
            # have no space there
            name = name.replace(" >", ">")
        if XML_AN_INCOMPLETE in attrs:
            decl = self.__decl_factory.create_class_declaration(name=name)
        else:
            decl = self.__decl_factory.create_class(
                name=name,
                class_type=class_type)
            decl.is_abstract = bool(attrs.get(XML_AN_ABSTRACT, False))
        self.__read_byte_size(decl, attrs)
        self.__read_byte_align(decl, attrs)
        if attrs.get(XML_AN_COMMENT) is not None:
            decl.comment = attrs.get(XML_AN_COMMENT)
        return decl

    def __read_class(self, attrs):
        return self.__read_class_impl(declarations.CLASS_TYPES.CLASS, attrs)

    def __read_struct(self, attrs):
        return self.__read_class_impl(declarations.CLASS_TYPES.STRUCT, attrs)

    def __read_union(self, attrs):
        return self.__read_class_impl(declarations.CLASS_TYPES.UNION, attrs)

    def __read_casting_operator(self, attrs):
        operator = self.__decl_factory.create_casting_operator()
        self.__read_member_function(operator, attrs, True)
        return operator

    def __read_comment(self, attrs):
        comment = self.__decl_factory.create_comment()
        comment.begin_line = int(attrs.get(XML_AN_BEGIN_LINE))
        comment.end_line = int(attrs.get(XML_AN_END_LINE))
        comment.begin_offset = int(attrs.get(XML_AN_BEGIN_OFFSET))
        comment.end_offset = int(attrs.get(XML_AN_END_OFFSET))
        comment.begin_column = int(attrs.get(XML_AN_BEGIN_COLUMN))
        comment.end_column = int(attrs.get(XML_AN_END_COLUMN))
        self.__read_location(comment, attrs, self.__name_attrs_to_skip)
        return comment

    def __read_constructor(self, attrs):
        constructor = self.__decl_factory.create_constructor()
        self.__read_member_function(constructor, attrs, True)
        constructor.explicit = attrs.get(XML_AN_EXPLICIT, False)
        return constructor

    def __read_function(self, attrs):
        gfunction = self.__decl_factory.create_free_function()
        self.__read_calldef(gfunction, attrs, True)
        return gfunction

    def __read_method(self, attrs):
        mfunction = self.__decl_factory.create_member_function()
        self.__read_member_function(mfunction, attrs, True)
        return mfunction

    def __read_destructor(self, attrs):
        destructor = self.__decl_factory.create_destructor()
        self.__read_member_function(destructor, attrs, True)
        destructor.name = '~' + destructor.name
        return destructor

    def __read_free_operator(self, attrs):
        operator = self.__decl_factory.create_free_operator()
        self.__read_member_function(operator, attrs, True)
        self.__update_operator_name(operator)
        return operator

    def __read_member_operator(self, attrs):
        operator = self.__decl_factory.create_member_operator()
        self.__read_member_function(operator, attrs, True)
        self.__update_operator_name(operator)
        return operator

    def __read_version(self, attrs):
        castxml_format = attrs.get(XML_AN_CASTXML_FORMAT)
        gccxml_cvs_revision = None
        if castxml_format is None:
            gccxml_cvs_revision = attrs.get(XML_AN_CVS_REVISION)
        xml_generator = utils.xml_generators(
            utils.loggers.cxx_parser, gccxml_cvs_revision, castxml_format)
        utils.xml_output_version = gccxml_cvs_revision
        self.__xml_generator_from_xml_file = xml_generator

    @staticmethod
    def __update_operator_name(operator):
        space = ""
        if "new" in operator.name or "delete" in operator.name:
            space = " "
        operator.name = "operator" + space + operator.name



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

from . import autoconfig

from pygccxml import parser
from pygccxml import declarations

TEST_FILES = [
    "null_comparison.hpp",
]


def test_argument_null_comparison():
    """
    Test for None comparisons with default arguments
    """

    config = autoconfig.cxx_parsers_cfg.config.clone()
    decls = parser.parse(TEST_FILES, config)
    global_ns = declarations.get_global_namespace(decls)

    ns = global_ns.namespace("ns")

    func = ns.free_function(name="TestFunction1")
    assert (func.arguments[0] > func.arguments[1]) is False

    func = ns.free_function(name="TestFunction2")
    assert (func.arguments[0] > func.arguments[1]) is False



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

import pytest

from . import autoconfig

from pygccxml import parser
from pygccxml import declarations


TEST_FILES = [
    "string_traits.hpp",
]


@pytest.fixture
def global_ns():
    COMPILATION_MODE = parser.COMPILATION_MODE.ALL_AT_ONCE
    config = autoconfig.cxx_parsers_cfg.config.clone()
    decls = parser.parse(TEST_FILES, config, COMPILATION_MODE)
    global_ns = declarations.get_global_namespace(decls)
    global_ns.init_optimizer()
    return global_ns


def validate_yes(ns, controller):
    for typedef in ns.typedefs():
        assert controller(typedef.decl_type) is True


def validate_no(ns, controller):
    for typedef in ns.typedefs():
        assert controller(typedef.decl_type) is False


def test_string(global_ns):
    string_traits = global_ns.namespace('string_traits')
    validate_yes(
        string_traits.namespace('yes'),
        declarations.is_std_string)
    validate_no(
        string_traits.namespace('no'),
        declarations.is_std_string)


def test_wstring(global_ns):
    wstring_traits = global_ns.namespace('wstring_traits')
    validate_yes(
        wstring_traits.namespace('yes'),
        declarations.is_std_wstring)
    validate_no(
        wstring_traits.namespace('no'),
        declarations.is_std_wstring)



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

import pytest

from . import autoconfig

from pygccxml import parser
from pygccxml import declarations


TEST_FILES = [
    "vector_traits.hpp",
]


@pytest.fixture
def global_ns():
    COMPILATION_MODE = parser.COMPILATION_MODE.ALL_AT_ONCE
    INIT_OPTIMIZER = True
    config = autoconfig.cxx_parsers_cfg.config.clone()
    # TOOD: this breaks the tests, check why
    # config.castxml_epic_version = 1
    decls = parser.parse(TEST_FILES, config, COMPILATION_MODE)
    global_ns = declarations.get_global_namespace(decls)
    if INIT_OPTIMIZER:
        global_ns.init_optimizer()
    return global_ns


def validate_yes(value_type, container):
    traits = declarations.vector_traits
    assert traits.is_my_case(container) is True
    assert declarations.is_same(
            value_type,
            traits.element_type(container)) is True
    assert traits.is_sequence(container) is True


def test_global_ns(global_ns):
    value_type = global_ns.class_('_0_')
    container = global_ns.typedef('container', recursive=False)
    validate_yes(value_type, container)


def test_yes(global_ns):
    yes_ns = global_ns.namespace('yes')
    for struct in yes_ns.classes():
        if not struct.name.startswith('_'):
            continue
        if not struct.name.endswith('_'):
            continue
        validate_yes(
            struct.typedef('value_type'),
            struct.typedef('container'))


def test_no(global_ns):
    traits = declarations.vector_traits
    no_ns = global_ns.namespace('no')
    for struct in no_ns.classes():
        if not struct.name.startswith('_'):
            continue
        if not struct.name.endswith('_'):
            continue
        assert traits.is_my_case(struct.typedef('container')) is False


def test_declaration():
    cnt = (
        'std::vector<std::basic_string<char, std::char_traits<char>, ' +
        'std::allocator<char>>,std::allocator<std::basic_string<char, ' +
        'std::char_traits<char>, std::allocator<char>>>>' +
        '@::std::vector<std::basic_string<char, std::char_traits<char>, ' +
        'std::allocator<char>>, std::allocator<std::basic_string<char, ' +
        'std::char_traits<char>, std::allocator<char>>>>')
    traits = declarations.find_container_traits(cnt)
    assert declarations.vector_traits == traits


def test_element_type(global_ns):
    do_nothing = global_ns.free_function('do_nothing')
    v = declarations.remove_reference(
        declarations.remove_declarated(
            do_nothing.arguments[0].decl_type))
    declarations.vector_traits.element_type(v)



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

from pygccxml import declarations


def test_extract():
    data = [
        ('thiscall',
            '(public: __thiscall std::auto_ptr<class pof::number_t>' +
            '::auto_ptr<class pof::number_t>(class std::auto_ptr' +
            '<class pof::number_t> &))'),
        ('',  "(const pof::number_t::`vftable')")]

    for expected, text in data:
        got = declarations.CALLING_CONVENTION_TYPES.extract(text)
        assert got == expected



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

import pytest

from . import autoconfig

from pygccxml import parser
from pygccxml import declarations

TEST_FILES = ["remove_template_defaults.hpp", "indexing_suites2.hpp"]


@pytest.fixture
def global_ns():
    COMPILATION_MODE = parser.COMPILATION_MODE.ALL_AT_ONCE
    config = autoconfig.cxx_parsers_cfg.config.clone()
    decls = parser.parse(TEST_FILES, config, COMPILATION_MODE)
    global_ns = declarations.get_global_namespace(decls)
    global_ns.init_optimizer()
    return global_ns


def __cmp_traits(global_ns, typedef, expected, partial_name, key_type=None):
    if isinstance(typedef, str):
        typedef = global_ns.typedef(typedef)
    traits = declarations.find_container_traits(typedef)
    assert traits is not None
    assert traits == expected
    cls = declarations.remove_declarated(typedef)
    assert declarations.find_container_traits(cls) == expected
    assert cls.partial_name == partial_name
    cls = traits.class_declaration(cls)
    assert traits.element_type(typedef) is not None
    assert cls.cache.container_element_type is not None

    if key_type:
        assert traits.is_mapping(typedef) is not None
        real_key_type = traits.key_type(typedef)
        assert real_key_type.decl_string == key_type
        assert cls.cache.container_key_type is not None
    else:
        assert traits.is_sequence(typedef)


def test_find_traits(global_ns):
    __cmp_traits(
        global_ns,
        "v_int",
        declarations.vector_traits,
        "vector<int>"
    )
    __cmp_traits(
        global_ns,
        "l_int",
        declarations.list_traits,
        "list<int>"
    )
    __cmp_traits(
        global_ns,
        "d_v_int",
        declarations.deque_traits,
        "deque<std::vector<int>>"
    )
    __cmp_traits(
        global_ns, "q_int",
        declarations.queue_traits,
        "queue<int>"
    )
    __cmp_traits(
        global_ns, "pq_int",
        declarations.priority_queue_traits,
        "priority_queue<int>"
    )
    __cmp_traits(
        global_ns, "s_v_int",
        declarations.set_traits,
        "set<std::vector<int>>"
    )
    __cmp_traits(
        global_ns,
        "ms_v_int",
        declarations.multiset_traits,
        "multiset<std::vector<int>>",
    )
    __cmp_traits(
        global_ns, "m_i2d",
        declarations.map_traits,
        "map<int, double>",
        "int"
    )
    __cmp_traits(
        global_ns,
        "mm_i2d",
        declarations.multimap_traits,
        "multimap<int, double>",
        "int",
    )
    __cmp_traits(
        global_ns,
        "hs_v_int",
        declarations.unordered_set_traits,
        "unordered_set<std::vector<int>>",
    )
    __cmp_traits(
        global_ns,
        "mhs_v_int",
        declarations.unordered_multiset_traits,
        "unordered_multiset<std::vector<int>>",
    )
    __cmp_traits(
        global_ns,
        "hm_i2d",
        declarations.unordered_map_traits,
        "unordered_map<int, double>",
        "int",
    )
    __cmp_traits(
        global_ns,
        "hmm_i2d",
        declarations.unordered_multimap_traits,
        "unordered_multimap<int, double>",
        "int",
    )


def test_multimap(global_ns):
    m = global_ns.class_(lambda decl: decl.name.startswith("multimap"))
    declarations.find_container_traits(m)
    assert m.partial_name == "multimap<int, int>"


def test_recursive_partial_name(global_ns):
    f1 = global_ns.free_function("f1")
    t1 = declarations.class_traits.get_declaration(f1.arguments[0].decl_type)
    assert "type<std::set<std::vector<int>>>" == t1.partial_name


def test_remove_defaults_partial_name_namespace(global_ns):
    f2 = global_ns.free_function("f2")
    type_info = f2.return_type
    traits = declarations.find_container_traits(type_info)
    cls = traits.class_declaration(type_info)
    decl_string = cls.partial_decl_string
    key_type_string = traits.key_type(type_info).partial_decl_string
    assert decl_string.startswith("::std::")
    assert key_type_string.startswith("::std::")


def test_from_ogre():
    x = (
        "map<std::string, bool (*)(std::string&, "
        + "Ogre::MaterialScriptContext&), std::less<std::string>, "
        + "std::allocator<std::pair<std::string const, bool (*)"
        + "(std::string&, Ogre::MaterialScriptContext&)>>>"
    )
    ct = declarations.find_container_traits(x)
    ct.remove_defaults(x)


def test_infinite_loop(global_ns):
    rt = global_ns.free_function("test_infinite_loop").return_type
    map_traits = declarations.find_container_traits(rt)
    assert map_traits == declarations.map_traits
    elem = map_traits.element_type(rt)
    assert elem.decl_string == "int"



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

import pytest

from . import autoconfig

from pygccxml import parser
from pygccxml import declarations

TEST_FILES = ['test_array_argument.hpp']


@pytest.fixture
def global_ns():
    COMPILATION_MODE = parser.COMPILATION_MODE.ALL_AT_ONCE
    config = autoconfig.cxx_parsers_cfg.config.clone()
    config.cflags = "-std=c++11"
    decls = parser.parse(TEST_FILES, config, COMPILATION_MODE)
    global_ns = declarations.get_global_namespace(decls)
    global_ns.init_optimizer()
    return global_ns


def test_array_argument(global_ns):

    """
    Test to ensure that function arguments' array size are kept intact
    rather than presented as pointers.

    """

    criteria = declarations.calldef_matcher(name="function")
    free_funcs = declarations.matcher.find(criteria, global_ns)
    for free_func in free_funcs:
        decl_string = free_func.create_decl_string(with_defaults=False)
        assert decl_string == "void ( ::test::* )( int [1024],int [512] )"
        arg1 = free_func.arguments[0]
        arg2 = free_func.arguments[1]
        assert arg1.decl_type.decl_string == "int [1024]"
        assert arg1.name == "arg1"
        assert declarations.type_traits.array_size(arg1.decl_type) == 1024
        assert isinstance(
            declarations.type_traits.array_item_type(arg1.decl_type),
            declarations.cpptypes.int_t
        ) is True
        assert arg2.decl_type.decl_string == "int [512]"
        assert arg2.name == "arg2"
        assert declarations.type_traits.array_size(arg2.decl_type) == 512
        assert isinstance(
            declarations.type_traits.array_item_type(arg2.decl_type),
            declarations.cpptypes.int_t
        ) is True



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

import pytest

import warnings

from . import autoconfig

from pygccxml import parser
from pygccxml import declarations


TEST_FILES = [
    "test_dynamic_exception.hpp",
]


@pytest.fixture
def global_ns():
    COMPILATION_MODE = parser.COMPILATION_MODE.ALL_AT_ONCE
    config = autoconfig.cxx_parsers_cfg.config.clone()
    # This test does not work with c++17 and above
    # See https://developers.redhat.com/articles/2021/08/06/porting-your-code-c17-gcc-11#exception_specification_changes # noqa
    # This test is thus excpected to use -std=c++14 forever
    config.cflags = "-std=c++14"
    decls = parser.parse(TEST_FILES, config, COMPILATION_MODE)
    global_ns = declarations.get_global_namespace(decls)
    global_ns.init_optimizer()
    return global_ns


def test_calldef_with_throw(global_ns, helpers):
    calldef_with_throw = global_ns.free_function("calldef_with_throw")
    assert calldef_with_throw is not None
    helpers._test_calldef_exceptions(
        global_ns, calldef_with_throw,
        ["some_exception_t", "other_exception_t"]
    )

    calldef_with_throw = global_ns.calldef('calldef_with_throw')
    some_exception = global_ns.class_('some_exception_t')
    other_exception = global_ns.class_('other_exception_t')

    # Legacy way of fetching dependencies. Is still valid but deprecated
    warnings.simplefilter("ignore", Warning)
    dependencies_old = calldef_with_throw.i_depend_on_them()
    warnings.simplefilter("error", Warning)
    assert len(dependencies_old) == 3
    dependencies_old = [
        dependency for dependency in dependencies_old if
        dependency.depend_on_it in (some_exception, other_exception)]
    assert len(dependencies_old) == 2

    dependencies_new = declarations.get_dependencies_from_decl(
        calldef_with_throw)
    assert len(dependencies_new) == 3
    dependencies_new = [
        dependency for dependency in dependencies_new if
        dependency.depend_on_it in (some_exception, other_exception)]
    assert len(dependencies_new) == 2



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

import pytest

import logging

from pygccxml import utils


mock_logger = logging.getLogger("Test")


def test_old_xml_generators():
    """
    Tests for the xml_generators class.

    This is for gccxml and for castxml using the gccxml xml file format
    """
    _test_impl("0.6", False, "is_gccxml_06")
    _test_impl("1.114", False, "is_gccxml_07")
    _test_impl("1.115", False, "is_gccxml_09_buggy")
    _test_impl("1.126", False, "is_gccxml_09_buggy")
    _test_impl("1.127", False, "is_gccxml_09")
    _test_impl("1.136", True, "is_castxml")


def test_casxtml_epic_version_1():
    """
    Test with the castxml epic version set to 1
    """
    gen = utils.xml_generators(
        mock_logger, castxml_format="1.1.0")
    assert gen.is_gccxml is False
    assert gen.is_castxml is True
    assert gen.is_castxml1 is True
    assert gen.xml_output_version == "1.1.0"

    with pytest.raises(RuntimeError):
        utils.xml_generators(mock_logger, "1.136", "1.1.0")

    with pytest.raises(RuntimeError):
        utils.xml_generators(mock_logger, None, None)


def _test_impl(
        gccxml_cvs_revision, is_castxml,
        expected_gccxml_cvs_revision):
    """
    Implementation detail for the test

    Args:
        gccxml_cvs_revision (str|None) : a known cvs revision
        is_castxml (bool): check for castxml
        expected_gccxml_cvs_revision (str): will be used to check if the
            attribute is set to True.
    """
    gen = utils.xml_generators(
        mock_logger, gccxml_cvs_revision)
    if is_castxml:
        assert gen.is_gccxml is False
        assert gen.is_castxml is True
    else:
        assert gen.is_gccxml is True
        assert gen.is_castxml is False
    assert getattr(gen, expected_gccxml_cvs_revision) is True
    assert gen.xml_output_version == gccxml_cvs_revision



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

import platform

import pytest

from . import autoconfig

from pygccxml import parser
from pygccxml import declarations
from pygccxml import utils

TEST_FILES = [
    'remove_template_defaults.hpp'
]


@pytest.fixture
def global_ns():
    COMPILATION_MODE = parser.COMPILATION_MODE.ALL_AT_ONCE
    config = autoconfig.cxx_parsers_cfg.config.clone()
    if platform.system() == "Darwin":
        config.cflags = "-std=c++11 -Dat_quick_exit=atexit -Dquick_exit=exit"
        # https://fr.mathworks.com/matlabcentral/answers/2013982-clibgen-generatelibrarydefinition-error-the-global-scope-has-no-quick_exit-on-mac-m2#answer_1439856
        # https://github.com/jetbrains/kotlin/commit/d50f585911dedec5723213da8835707ac95e1c01
    else:
        config.cflags = "-std=c++11"
    decls = parser.parse(TEST_FILES, config, COMPILATION_MODE)
    global_ns = declarations.get_global_namespace(decls)
    global_ns.init_optimizer()
    return global_ns


def test_vector(global_ns):
    v_int = global_ns.typedef('v_int')
    v_traits = declarations.vector_traits
    assert 'vector<int>' == v_traits.remove_defaults(v_int)
    v_string = global_ns.typedef('v_string')
    assert 'vector<std::string>' == \
        v_traits.remove_defaults(v_string)
    v_v_int = global_ns.typedef('v_v_int')
    assert 'vector<std::vector<int>>' == \
        v_traits.remove_defaults(v_v_int)


def test_list(global_ns):
    l_int = global_ns.typedef('l_int')
    l_traits = declarations.list_traits
    assert 'list<int>' == l_traits.remove_defaults(l_int)
    l_wstring = global_ns.typedef('l_wstring')
    assert 'list<std::wstring>' == l_traits.remove_defaults(l_wstring)


def test_deque(global_ns):
    d_v_int = global_ns.typedef('d_v_int')
    d_v_traits = declarations.deque_traits
    assert 'deque<std::vector<int>>' == \
        d_v_traits.remove_defaults(d_v_int)
    d_l_string = global_ns.typedef('d_l_string')
    assert 'deque<std::list<std::string>>' == \
        d_v_traits.remove_defaults(d_l_string)


def test_queue(global_ns):
    q_int = global_ns.typedef('q_int')
    q_traits = declarations.queue_traits
    assert 'queue<int>' == q_traits.remove_defaults(q_int)
    q_string = global_ns.typedef('q_string')
    assert 'queue<std::string>' == q_traits.remove_defaults(q_string)


def test_priority_queue(global_ns):
    pq_int = global_ns.typedef('pq_int')
    pq_traits = declarations.priority_queue_traits
    assert 'priority_queue<int>' == pq_traits.remove_defaults(pq_int)
    pq_string = global_ns.typedef('pq_string')
    assert 'priority_queue<std::string>' == \
        pq_traits.remove_defaults(pq_string)


def test_set(global_ns):
    s_v_int = global_ns.typedef('s_v_int')
    assert 'set<std::vector<int>>' == \
        declarations.set_traits.remove_defaults(s_v_int)
    s_string = global_ns.typedef('s_string')
    assert 'set<std::string>' == \
        declarations.set_traits.remove_defaults(s_string)


def test_multiset(global_ns):
    ms_v_int = global_ns.typedef('ms_v_int')
    ms_v_traits = declarations.multiset_traits
    assert 'multiset<std::vector<int>>' == \
        ms_v_traits.remove_defaults(ms_v_int)
    ms_string = global_ns.typedef('ms_string')
    assert 'multiset<std::string>' == \
        ms_v_traits.remove_defaults(ms_string)


def test_map(global_ns):
    m_i2d = global_ns.typedef('m_i2d')
    assert 'map<int, double>' == \
        declarations.map_traits.remove_defaults(m_i2d)
    m_wstr2d = global_ns.typedef('m_wstr2d')
    assert 'map<std::wstring, double>' == \
        declarations.map_traits.remove_defaults(m_wstr2d)
    m_v_i2m_wstr2d = global_ns.typedef('m_v_i2m_wstr2d')
    m = 'map<const std::vector<int>, std::map<std::wstring, double>>'
    assert m == declarations.map_traits.remove_defaults(m_v_i2m_wstr2d)


def test_multimap(global_ns):
    mm_i2d = global_ns.typedef('mm_i2d')
    mm_traits = declarations.multimap_traits
    assert 'multimap<int, double>' == mm_traits.remove_defaults(mm_i2d)
    mm_wstr2d = global_ns.typedef('mm_wstr2d')
    assert 'multimap<const std::wstring, double>' == \
        mm_traits.remove_defaults(mm_wstr2d)
    mm_v_i2mm_wstr2d = global_ns.typedef('mm_v_i2mm_wstr2d')
    assert ('multimap<const std::vector<int>, ' +
            'const std::multimap<const std::wstring, double>>') == \
        mm_traits.remove_defaults(mm_v_i2mm_wstr2d)


def test_hash_set(global_ns):
    hs_v_int = global_ns.typedef('hs_v_int')
    hs_traits = declarations.unordered_set_traits
    name = 'unordered_set'
    assert (name + '<std::vector<int>>') == \
        hs_traits.remove_defaults(hs_v_int), \
        hs_traits.remove_defaults(hs_v_int)
    hs_string = global_ns.typedef('hs_string')
    assert (name + '<std::string>') == \
        hs_traits.remove_defaults(hs_string)


def test_hash_multiset(global_ns):
    mhs_v_int = global_ns.typedef('mhs_v_int')
    mhs_traits = declarations.unordered_multiset_traits
    name = 'unordered_multiset'
    assert (name + '<std::vector<int>>') == \
        mhs_traits.remove_defaults(mhs_v_int)
    mhs_string = global_ns.typedef('mhs_string')
    assert (name + '<std::string>') == \
        mhs_traits.remove_defaults(mhs_string)


def test_hash_map(global_ns):
    hm_i2d = global_ns.typedef('hm_i2d')
    hm_traits = declarations.unordered_map_traits
    name = 'unordered_map'
    assert (name + '<int, double>') == \
        hm_traits.remove_defaults(hm_i2d)
    hm_wstr2d = global_ns.typedef('hm_wstr2d')
    assert (name + '<std::wstring, double>') == \
        hm_traits.remove_defaults(hm_wstr2d)


def test_hash_multimap(global_ns):
    hmm_i2d = global_ns.typedef('hmm_i2d')
    hmm_traits = declarations.unordered_multimap_traits
    name = 'unordered_multimap'
    assert (name + '<int, double>') == \
        hmm_traits.remove_defaults(hmm_i2d)
    hmm_wstr2d = global_ns.typedef('hmm_wstr2d')
    assert (name + '<const std::wstring, double>') == \
        hmm_traits.remove_defaults(hmm_wstr2d)

    hmm_v_i2mm_wstr2d = global_ns.typedef('hmm_v_i2mm_wstr2d')

    hmm_traits_value = hmm_traits.remove_defaults(hmm_v_i2mm_wstr2d)

    possible_values = (
        name + '<const std::vector<int>, ' +
        'const __gnu_cxx::' + name + '<const std::wstring, double>>',
        name + '<const std::vector<int>, ' +
        'const std::' + utils.get_tr1(hmm_traits_value) + name +
        '<const std::wstring, double>>',
        name + '<const std::vector<int>, ' +
        'const stdext::' + name + '<const std::wstring, double>>')

    assert hmm_traits_value in possible_values, hmm_traits_value



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

from pygccxml import utils
from pygccxml import declarations
from pygccxml import parser

import os
import sys
import warnings
warnings.simplefilter("error", Warning)
# Find out the file location within the sources tree
this_module_dir_path = os.path.abspath(
    os.path.dirname(sys.modules[__name__].__file__))

# Find out the c++ parser
generator_path, generator_name = utils.find_xml_generator()

# Configure the xml generator
xml_generator_config = parser.xml_generator_configuration_t(
    xml_generator_path=generator_path,
    xml_generator=generator_name)

# The c++ file we want to parse
filename = "example.hpp"
filename = this_module_dir_path + "/" + filename

decls = parser.parse([filename], xml_generator_config)
global_namespace = declarations.get_global_namespace(decls)
ns = global_namespace.namespace("ns")

class_t_decl = []
for d in ns.declarations:
    if isinstance(d, declarations.class_declaration_t):
        class_declaration_t = d
    if isinstance(d, declarations.class_t):
        class_t_decl.append(d)
    if isinstance(d, declarations.free_function_t):
        free_function_t_decl = d

print(class_t_decl[0])
# > ns::B [struct]

print(class_t_decl[1])
# > ns::D [struct]

print(class_declaration_t)
# > ns::T<ns::B::D, bool> [class declaration]

print(free_function_t_decl)
# > ns::T<ns::B::D, bool> ns::function() [free function]

print(declarations.templates.is_instantiation(class_declaration_t.name))
# > True

name, parameter_list = declarations.templates.split(class_declaration_t.name)
print(name, parameter_list)
# > 'T', ['ns::B::D', 'bool']



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

import pytest

from . import autoconfig

from pygccxml import parser


def test_castxml_epic_version_check():
    """
    Test using a forbidden value for the castxml epic version.

    """

    config = autoconfig.cxx_parsers_cfg.config.clone()
    config.castxml_epic_version = 2
    with pytest.raises(RuntimeError):
        parser.parse_string("", config)



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

import pytest

from . import autoconfig


from pygccxml import parser
from pygccxml import declarations

TEST_FILES = "core_ns_join_1.hpp"


def __check_result(decls):
    E11 = declarations.find_declaration(decls, fullname='::E11')
    assert E11 is not None
    ns12 = declarations.find_declaration(decls, fullname='::ns::ns12')
    assert ns12 is not None
    E13 = declarations.find_declaration(ns12.declarations, name='E13')
    assert E13 is not None
    E14 = declarations.find_declaration(decls, name='E14')
    assert E14 is None


def test_simple_start_with_declarations():
    config = autoconfig.cxx_parsers_cfg.config.clone()
    config.start_with_declarations.extend(['E11', 'ns::ns12::E13'])
    decls = parser.parse([TEST_FILES], config)
    __check_result(decls)


def test_project_reader_file_by_file_start_with_declarations():
    config = autoconfig.cxx_parsers_cfg.config.clone()
    config.start_with_declarations.extend(['E11', 'ns::ns12::E13'])
    reader = parser.project_reader_t(config)
    decls = reader.read_files(
        [parser.file_configuration_t(
            TEST_FILES, config.start_with_declarations)],
        parser.COMPILATION_MODE.FILE_BY_FILE)
    __check_result(decls)


def test_project_reader_all_at_once_start_with_declarations():
    config = autoconfig.cxx_parsers_cfg.config.clone()
    config.start_with_declarations.extend(['E11', 'ns::ns12::E13'])
    reader = parser.project_reader_t(config)
    decls = reader.read_files(
        [parser.file_configuration_t(
            TEST_FILES, config.start_with_declarations)],
        parser.COMPILATION_MODE.ALL_AT_ONCE)
    __check_result(decls)



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

import pytest

from . import autoconfig

from pygccxml import declarations
from pygccxml import parser
from pygccxml import utils


TEST_FILES = [
    "patcher.hpp",
]

config = autoconfig.cxx_parsers_cfg.config.clone()
project_reader = parser.project_reader_t(config=config, cache=None)
decls = project_reader.read_files(
    TEST_FILES,
    parser.COMPILATION_MODE.ALL_AT_ONCE
)
config.xml_generator_from_xml_file = project_reader.xml_generator_from_xml_file
config.__cxx_std = utils.cxx_standard(config.cflags)


@pytest.fixture
def global_ns():
    global_ns = declarations.get_global_namespace(decls)
    return global_ns


def test_enum_patcher(global_ns):
    fix_enum = global_ns.free_function("fix_enum")
    default_val = fix_enum.arguments[0].default_value
    if config.__cxx_std.is_cxx11_or_greater:
        val = "::ns1::ns2::fruit::apple"
    else:
        val = "::ns1::ns2::apple"
    assert default_val == val

    fix_enum2 = global_ns.free_function("fix_enum2")
    default_val = fix_enum2.arguments[0].default_value
    assert default_val == val

    ns1 = global_ns.namespace("ns1")
    ns2 = ns1.namespace("ns2")
    fix_enum2 = ns2.free_function("fix_enum2")
    default_val = fix_enum2.arguments[0].default_value
    assert default_val == val

    fix_enum3 = global_ns.free_function("fix_enum3")
    default_val = fix_enum3.arguments[0].default_value
    val = val.replace("apple", "orange")
    assert default_val == val

    if config.__cxx_std.is_cxx11_or_greater:
        fix_enum4 = global_ns.free_function("fix_enum4")
        default_val = fix_enum4.arguments[0].default_value
        assert default_val == "::ns4::color::blue"

        fix_enum5 = global_ns.free_function("fix_enum5")
        default_val = fix_enum5.arguments[0].default_value
        assert default_val == "::ns4::color::blue"

    lpe = global_ns.free_function("log_priority_enabled")
    default_val = lpe.arguments[0].default_value
    if config.__cxx_std.is_cxx11_or_greater:
        val = "(long int)" + \
            "(::ACE_Log_Priority_Index::LM_INVALID_BIT_INDEX)"
    else:
        val = "(long int)(::LM_INVALID_BIT_INDEX)"
    assert default_val == val


def test_numeric_patcher(global_ns):
    fix_numeric = global_ns.free_function("fix_numeric")
    generator = config.xml_generator_from_xml_file
    if generator.is_castxml1 or \
            float(generator.xml_output_version) >= 1.137:
        val = "(unsigned long long)-1"
    else:
        val = "(ull)-1"
    assert fix_numeric.arguments[0].default_value == val


def test_unqualified_integral_patcher(global_ns):
    # For this check to be removed, patcher_tester_64bit.xml
    # will need to be updated for CastXML
    return

    ns1 = global_ns.namespace("ns1")
    st1 = ns1.class_("st1")
    fun1 = st1.member_function("fun1")
    output_verion = xml_generator_from_xml_file.xml_output_version
    if xml_generator_from_xml_file.is_castxml1 or \
            float(output_verion) >= 1.137:
        val1 = "ns1::DEFAULT_1"
        val2 = "ns1::st1::DEFAULT_2"
    else:
        val1 = "::ns1::DEFAULT_1"
        val2 = "::ns1::st1::DEFAULT_2"
    assertEqual(
        fun1.arguments[0].default_value, val1)
    assertEqual(
        fun1.arguments[1].default_value, val2)

    fun2 = global_ns.free_function("fun2")
    assertEqual(
        fun2.arguments[0].default_value,
        "::DEFAULT_1")
    output_verion = xml_generator_from_xml_file.xml_output_version
    if xml_generator_from_xml_file.is_castxml1 or \
            float(output_verion) >= 1.137:
        val1 = "ns1::DEFAULT_1"
        val2 = "ns1::st1::DEFAULT_2"
    else:
        # Before XML output version 1.137, the following two
        # were unpatched and were identical to the text in
        # matcher.hpp
        val1 = "ns1::DEFAULT_1"
        val2 = "::ns1::st1::DEFAULT_2"
    assertEqual(
        fun2.arguments[1].default_value, val1)
    assertEqual(
        fun2.arguments[2].default_value, val2)


def test_unnamed_enum_patcher(global_ns):
    fix_unnamed = global_ns.free_function("fix_unnamed")
    assert fix_unnamed.arguments[0].default_value == "int(::fx::unnamed)"


def test_function_call_patcher(global_ns):
    fix_function_call = global_ns.free_function("fix_function_call")
    default_val = fix_function_call.arguments[0].default_value
    output_verion = config.xml_generator_from_xml_file.xml_output_version
    if config.xml_generator_from_xml_file.is_castxml1 or \
            float(output_verion) >= 1.137:
        val = "function_call::calc(1, 2, 3)"
    else:
        val = "calc(1, 2, 3)"
    assert default_val == val


def test_fundamental_patcher(global_ns):
    fcall = global_ns.free_function("fix_fundamental")
    if config.__cxx_std.is_cxx11_or_greater:
        val = "(unsigned int)(::fundamental::spam::eggs)"
    else:
        val = "(unsigned int)(::fundamental::eggs)"
    assert fcall.arguments[0].default_value == val


def test_constructor_patcher(global_ns):
    typedef__func = global_ns.free_function("typedef__func")
    default_val = typedef__func.arguments[0].default_value
    val = "typedef_::alias()"
    assert default_val == val



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

import pytest

from . import autoconfig

from pygccxml import parser
from pygccxml import declarations


TEST_FILES = [
    "type_as_exception_bug.h",
]


@pytest.fixture
def global_ns():
    COMPILATION_MODE = parser.COMPILATION_MODE.ALL_AT_ONCE
    config = autoconfig.cxx_parsers_cfg.config.clone()
    # This test does not work with c++17 and above
    # See https://developers.redhat.com/articles/2021/08/06/porting-your-code-c17-gcc-11#exception_specification_changes # noqa
    # This test is thus excpected to use -std=c++14 forever
    config.cflags = "-std=c++14"
    decls = parser.parse(TEST_FILES, config, COMPILATION_MODE)
    global_ns = declarations.get_global_namespace(decls)
    return global_ns


def test_type_as_exception(global_ns):
    buggy = global_ns.member_function('buggy')
    expression_error = global_ns.class_('ExpressionError')
    assert len(buggy.exceptions) == 1
    err = buggy.exceptions[0]
    assert declarations.is_reference(err)
    err = declarations.remove_declarated(
        declarations.remove_reference(err))
    assert err is expression_error



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

import pytest

from . import autoconfig

from pygccxml import parser
from pygccxml import declarations


TEST_FILES = ['core_membership.hpp']


@pytest.fixture
def global_ns():
    COMPILATION_MODE = parser.COMPILATION_MODE.ALL_AT_ONCE
    config = autoconfig.cxx_parsers_cfg.config.clone()
    decls = parser.parse(TEST_FILES, config, COMPILATION_MODE)
    global_ns = declarations.get_global_namespace(decls)
    global_ns.init_optimizer()
    return global_ns


def test_name_based(global_ns):
    cls = global_ns.class_(name='class_for_nested_enums_t')

    cls_full_name = declarations.full_name(cls)
    assert cls.cache.full_name == cls_full_name

    cls_declaration_path = declarations.declaration_path(cls)
    assert cls.cache.declaration_path == cls_declaration_path

    enum = cls.enumeration('ENestedPublic')

    enum_full_name = declarations.full_name(enum)
    assert enum.cache.full_name == enum_full_name

    enum_declaration_path = declarations.declaration_path(enum)
    assert enum.cache.declaration_path == enum_declaration_path

    # now we change class name, all internal decls cache should be cleared
    cls.name = "new_name"
    assert not cls.cache.full_name
    assert not cls.cache.declaration_path

    assert not enum.cache.full_name
    assert not enum.cache.declaration_path


def test_access_type(global_ns):
    cls = global_ns.class_(name='class_for_nested_enums_t')
    enum = cls.enumeration('ENestedPublic')
    assert enum.cache.access_type == 'public'
    enum.cache.reset_access_type()
    assert not enum.cache.access_type
    assert 'public' == cls.find_out_member_access_type(enum)
    assert enum.cache.access_type == 'public'



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

import copy

from . import autoconfig

from pygccxml import parser
from pygccxml import declarations

TEST_FILES = [
    "declarations_comparison.hpp",
]


def test_comparison_declaration_by_declaration():
    config = autoconfig.cxx_parsers_cfg.config.clone()
    config.castxml_epic_version = 1
    parsed = parser.parse(TEST_FILES, config)
    copied = copy.deepcopy(parsed)
    parsed = declarations.make_flatten(parsed)
    copied = declarations.make_flatten(copied)
    parsed.sort()
    copied.sort()
    failuers = []
    for parsed_decl, copied_decl, index in \
            zip(parsed, copied, list(range(len(copied)))):

        if parsed_decl != copied_decl:
            failuers.append(
                ("__lt__ and/or __qe__ does not working " +
                    "properly in case of %s, %s, index %d") %
                (parsed_decl.__class__.__name__,
                    copied_decl.__class__.__name__, index))
    assert failuers == []


def test_comparison_from_reverse():
    config = autoconfig.cxx_parsers_cfg.config.clone()
    config.castxml_epic_version = 1
    parsed = parser.parse(TEST_FILES, config)
    copied = copy.deepcopy(parsed)
    parsed.sort()
    copied.reverse()
    copied.sort()
    x = parsed[4:6]
    x.sort()
    y = copied[4:6]
    y.sort()
    assert parsed == copied


def test___lt__transitivnost():
    ns_std = declarations.namespace_t(name='std')
    ns_global = declarations.namespace_t(name='::')
    ns_internal = declarations.namespace_t(name='ns')
    ns_internal.parent = ns_global
    ns_global.declarations.append(ns_internal)
    left2right = [ns_std, ns_global]
    right2left = [ns_global, ns_std]
    left2right.sort()
    right2left.sort()
    assert left2right == right2left


def test_same_declarations_different_intances():
    config = autoconfig.cxx_parsers_cfg.config.clone()
    config.castxml_epic_version = 1
    parsed = parser.parse(TEST_FILES, config)
    copied = copy.deepcopy(parsed)
    assert parsed == copied



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

"""
Implementation details

"""

import re


class parser_t(object):

    """implementation details"""

    def __init__(
            self,
            pattern_char_begin,
            pattern_char_end,
            pattern_char_separator):
        self.__begin = pattern_char_begin
        self.__end = pattern_char_end
        self.__separator = pattern_char_separator
        # right now parser does not take into account next qualifiers, but it
        # will
        self.__text_qualifier = '"'
        self.__char_qualifier = "'"
        self.__escape = '\\'

    def has_pattern(self, decl_string):
        """
        Implementation detail

        """
        if self.__begin == "<":
            # Cleanup parentheses blocks before checking for the pattern
            # See also the args() method (in this file) for more explanations.
            decl_string = re.sub("\\s\\(.*?\\)", "", decl_string).strip()

        last_part = decl_string.split('::')[-1]
        return (
            decl_string.find(self.__begin) != -1 and
            last_part.find(self.__end) != -1
        )

    def name(self, decl_string):
        """implementation details"""
        if not self.has_pattern(decl_string):
            return decl_string
        args_begin = decl_string.find(self.__begin)
        return decl_string[0: args_begin].strip()

    def __find_args_separator(self, decl_string, start_pos):
        """implementation details"""
        bracket_depth = 0
        for index, ch in enumerate(decl_string[start_pos:]):
            if ch not in (self.__begin, self.__end, self.__separator):
                continue  # I am interested only in < and >
            elif self.__separator == ch:
                if not bracket_depth:
                    return index + start_pos
            elif self.__begin == ch:
                bracket_depth += 1
            elif not bracket_depth:
                return index + start_pos
            else:
                bracket_depth -= 1
        return -1

    def args(self, decl_string):
        """
        Extracts a list of arguments from the provided declaration string.

        Implementation detail. Example usages:
        Input: myClass<std::vector<int>, std::vector<double>>
        Output: [std::vector<int>, std::vector<double>]

        Args:
           decl_string (str): the full declaration string

        Returns:
            list: list of arguments as strings

        """
        args_begin = decl_string.find(self.__begin)
        args_end = decl_string.rfind(self.__end)
        if -1 in (args_begin, args_end) or args_begin == args_end:
            raise RuntimeError(
                "%s doesn't validate template instantiation string" %
                decl_string)

        args_only = decl_string[args_begin + 1: args_end].strip()

        # The list of arguments to be returned
        args = []

        parentheses_blocks = []
        prev_span = 0
        if self.__begin == "<":
            # In case where we are splitting template names, there
            # can be parentheses blocks (for arguments) that need to be taken
            # care of.

            # Build a regex matching a space (\s)
            # + something inside parentheses
            regex = re.compile("\\s\\(.*?\\)")
            for m in regex.finditer(args_only):
                # Store the position and the content
                parentheses_blocks.append([m.start() - prev_span, m.group()])
                prev_span = m.end() - m.start()
                # Cleanup the args_only string by removing the parentheses and
                # their content.
                args_only = args_only.replace(m.group(), "")

        # Now we are trying to split the args_only string in multiple arguments
        previous_found, found = 0, 0
        while True:
            found = self.__find_args_separator(args_only, previous_found)
            if found == -1:
                args.append(args_only[previous_found:].strip())
                # This is the last argument. Break out of the loop.
                break
            else:
                args.append(args_only[previous_found: found].strip())
            previous_found = found + 1  # skip found separator

        # Get the size and position for each argument
        absolute_pos_list = []
        absolute_pos = 0
        for arg in args:
            absolute_pos += len(arg)
            absolute_pos_list.append(absolute_pos)

        for item in parentheses_blocks:
            # In case where there are parentheses blocks we add them back
            # to the right argument
            parentheses_block_absolute_pos = item[0]
            parentheses_block_string = item[1]

            current_arg_absolute_pos = 0
            for arg_index, arg_absolute_pos in enumerate(absolute_pos_list):
                current_arg_absolute_pos += arg_absolute_pos
                if current_arg_absolute_pos >= parentheses_block_absolute_pos:
                    # Add the parentheses block back and break out of the loop.
                    args[arg_index] += parentheses_block_string
                    break

        return args

    NOT_FOUND = (-1, -1)
    """implementation details"""

    def find_args(self, text, start=None):
        """implementation details"""
        if start is None:
            start = 0
        first_occurance = text.find(self.__begin, start)
        if first_occurance == -1:
            return self.NOT_FOUND
        previous_found, found = first_occurance + 1, 0
        while True:
            found = self.__find_args_separator(text, previous_found)
            if found == -1:
                return self.NOT_FOUND
            elif text[found] == self.__end:
                return first_occurance, found
            else:
                previous_found = found + 1  # skip found sep

    def split(self, decl_string):
        """implementation details"""
        assert self.has_pattern(decl_string)
        return self.name(decl_string), self.args(decl_string)

    def split_recursive(self, decl_string):
        """implementation details"""
        assert self.has_pattern(decl_string)
        to_go = [decl_string]
        while to_go:
            name, args = self.split(to_go.pop())
            for arg in args:
                if self.has_pattern(arg):
                    to_go.append(arg)
            yield name, args

    def join(self, name, args, arg_separator=None):
        """implementation details"""
        if None is arg_separator:
            arg_separator = ', '
        args = [_f for _f in args if _f]

        if not args:
            args_str = ''
        elif len(args) == 1:
            args_str = args[0]
        else:
            args_str = arg_separator.join(args)

        return ''.join([name, self.__begin, args_str, self.__end])

    def normalize(self, decl_string, arg_separator=None):
        """implementation details"""
        if not self.has_pattern(decl_string):
            return decl_string
        name, args = self.split(decl_string)
        for i, arg in enumerate(args):
            args[i] = self.normalize(arg)
        return self.join(name, args, arg_separator)



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

"""
defines all "built-in" classes that implement declarations compare
functionality according to some criteria
"""

import re
from . import class_declaration


class matcher_base_t(object):

    """matcher_base_t class defines interface for classes that will implement
       compare functionality according to some criteria.
    """

    def __init__(self):
        object.__init__(self)

    def __call__(self, decl):
        raise NotImplementedError(
            "matcher must always implement the __call__() method.")

    def __invert__(self):
        """not-operator (~)"""
        return not_matcher_t(self)

    def __and__(self, other):
        """and-operator (&)"""
        return and_matcher_t([self, other])

    def __or__(self, other):
        """or-operator (|)"""
        return or_matcher_t([self, other])

    def __str__(self):
        return "base class for all matchers"


class and_matcher_t(matcher_base_t):

    """
    Combine several other matchers with "&" (and) operator.

    For example: find all private functions with name XXX

    .. code-block:: python

       matcher = access_type_matcher_t( 'private' ) & \
           calldef_matcher_t( name='XXX' )
    """

    def __init__(self, matchers):
        matcher_base_t.__init__(self)
        self.matchers = matchers

    def __call__(self, decl):
        for matcher in self.matchers:
            if not matcher(decl):
                return False
        return True

    def __str__(self):
        return " & ".join(["(%s)" % str(x) for x in self.matchers])


class or_matcher_t(matcher_base_t):

    """Combine several other matchers with "|" (or) operator.

    For example: find all functions and variables with name 'XXX'

    .. code-block:: python

       matcher = variable_matcher_t( name='XXX' ) | \
           calldef_matcher_t( name='XXX' )

    """

    def __init__(self, matchers):
        matcher_base_t.__init__(self)
        self.matchers = matchers

    def __call__(self, decl):
        for matcher in self.matchers:
            if matcher(decl):
                return True
        return False

    def __str__(self):
        return " | ".join(["(%s)" % str(x) for x in self.matchers])


class not_matcher_t(matcher_base_t):

    """
    return the inverse result of a matcher

    For example: find all public and protected declarations

    .. code-block:: python

       matcher = ~access_type_matcher_t( 'private' )
    """

    def __init__(self, matcher):
        matcher_base_t.__init__(self)
        self.matcher = matcher

    def __call__(self, decl):
        return not self.matcher(decl)

    def __str__(self):
        return "~(%s)" % str(self.matcher)


class regex_matcher_t(matcher_base_t):

    """
    Instance of this class will match declaration using regular expression.
    User should supply a function that will extract from declaration desired
    information as string. Later, this matcher will match that string using
    user regular expression.
    """

    def __init__(self, regex, function=None):
        """
        :param regex: regular expression
        :type regex: string, an instance of this class will compile it for you

        :param function: function that will be called to get an information
                         from declaration as string. As input this function
                         takes single argument - reference to a declaration.
                         Return value should be string. If function is None,
                         then the matcher will use declaration name.

        """
        matcher_base_t.__init__(self)
        self.regex = re.compile(regex)
        self.function = function
        if self.function is None:
            self.function = lambda decl: decl.name

    def __call__(self, decl):
        text = self.function(decl)
        return bool(self.regex.match(text))

    def __str__(self):
        return '(regex=%s)' % self.regex


class custom_matcher_t(matcher_base_t):

    """
    Instance of this class will match declaration by user custom criteria.
    """

    def __init__(self, function):
        """
        :param function: callable, that takes single argument -
                         declaration instance should return True or False
        """
        matcher_base_t.__init__(self)
        self.function = function

    def __call__(self, decl):
        return bool(self.function(decl))

    def __str__(self):
        return '(user criteria)'


class access_type_matcher_t(matcher_base_t):

    """
    Instance of this class will match declaration by its access type: public,
    private or protected. If declarations does not have access type, for
    example free function, then `False` will be returned.
    """

    def __init__(self, access_type):
        """
        :param access_type: declaration access type, could be "public",
        "private", "protected"
        :type access_type: :class: `str`
        """
        matcher_base_t.__init__(self)
        self.access_type = access_type

    def __call__(self, decl):
        if not isinstance(decl.parent, class_declaration.class_t):
            return False
        return (
            self.access_type == decl.parent.find_out_member_access_type(decl)
        )

    def __str__(self):
        return '(access type=%s)' % self.access_type


class virtuality_type_matcher_t(matcher_base_t):

    """
    Instance of this class will match declaration by its virtual type: not
    virtual, virtual or pure virtual. If declarations does not have "virtual"
    property, for example free function, then `False` will be returned.
    """

    def __init__(self, virtuality_type):
        """
        :param access_type: declaration access type
        :type access_type: :class:VIRTUALITY_TYPES defines few constants for
        your convenience.
        """
        matcher_base_t.__init__(self)
        self.virtuality_type = virtuality_type

    def __call__(self, decl):
        if not isinstance(decl.parent, class_declaration.class_t):
            return False
        return self.virtuality_type == decl.virtuality

    def __str__(self):
        return '(virtuality type=%s)' % self.virtuality_type
